class Mammal:
    def __init__(self, species):
        self.species = species

    def __eq_(self, o):
        return hash(o.species) == hash(self.species)

    def __hash__(self):
        return hash(self.species)

    def __repr__(self):
        return "<" + self.species + ">|"

    def __str__(self):
        return "<" + self.species + ">|"


class MockWritable:
    def __init__(self):
        self.data = None

    def write(self, data):
        self.data = "write:" + data

    def flush(self):
        self.data += "flush"

    def close(self):
        self.data += "close"


class MockCallable:
    def __init__(self):
        self.data = None

    def __call__(self, data):
        self.data = "call:" + data


class WritableCallable(MockWritable, MockCallable):
    pass



import sys
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Generator,
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np

from pocketutils.core._internal import nicesize
from pocketutils.core.exceptions import RefusingRequestError
from pocketutils.core.input_output import DevNull
from pocketutils.tools.base_tools import BaseTools

Y = TypeVar("Y")
T = TypeVar("T")
Z = TypeVar("Z")
Q = TypeVar("Q")


class CommonTools(BaseTools):
    @classmethod
    def limit(cls, items: Iterable[Q], n: int) -> Generator[Q, None, None]:
        for i, x in zip(range(n), items):
            yield x

    @classmethod
    def try_none(
        cls,
        function: Callable[[], T],
        fail_val: Optional[T] = None,
        exception=Exception,
    ) -> Optional[T]:
        """
        Returns the value of a function or None if it raised an exception.

        Args:
            function: Try calling this function
            fail_val: Return this value
            exception: Restrict caught exceptions to subclasses of this type
        """
        # noinspection PyBroadException
        try:
            return function()
        except exception:
            return fail_val

    @classmethod
    def succeeds(cls, function: Callable[[], Any], exception=Exception) -> bool:
        """Returns True iff ``function`` does not raise an error."""
        return cls.try_none(function, exception=exception) is not None

    @classmethod
    def or_null(cls, x: Any, dtype=lambda s: s, or_else: Any = None) -> Optional[Any]:
        """
        Return ``None`` if the operation ``dtype`` on ``x`` failed; returns the result otherwise.
        """
        return or_else if cls.is_null(x) else dtype(x)

    @classmethod
    def or_raise(
        cls,
        x: Any,
        dtype=lambda s: s,
        or_else: Union[None, BaseException, Type[BaseException]] = None,
    ) -> Any:
        """
        Returns ``dtype(x)`` if ``x`` is not None, or raises ``or_else``.
        """
        if or_else is None:
            or_else = LookupError(f"Value is {x}")
        elif isinstance(or_else, type):
            or_else = or_else(f"Value is {x}")
        if cls.is_null(x):
            raise or_else
        return dtype(x)

    @classmethod
    def iterator_has_elements(cls, x: Iterator[Any]) -> bool:
        """
        Returns False iff ``next(x)`` raises a ``StopIteration``.
        WARNING: Tries to call ``next(x)``, progressing iterators. Don't use ``x`` after calling this.
        Note that calling ``iterator_has_elements([5])`` will raise a `TypeError`

        Args:
            x: Must be an Iterator
        """
        return cls.succeeds(lambda: next(x), StopIteration)

    @classmethod
    def is_null(cls, x: Any) -> bool:
        """
        Returns True if x is either:
            - None
            - NaN
        """
        try:
            if np.isnan(x):
                return True
        except (ValueError, TypeError):
            pass
        return x is None or x == float("nan")

    @classmethod
    def is_empty(cls, x: Any) -> bool:
        """
        Returns True iff either:
            - x is None
            - np.is_nan(x)
            - x is something with 0 length
            - x is iterable and has 0 elements (will call ``__iter__``)

        Raises:
            RefusingRequestError If ``x`` is an Iterator. Calling this would empty the iterator, which is dangerous.
        """
        if isinstance(x, Iterator):
            raise RefusingRequestError("Do not call is_empty on an iterator.")
        try:
            if np.isnan(x):
                return True
        except (ValueError, TypeError):
            pass
        return (
            x is None
            or x == float("nan")
            or hasattr(x, "__len__")
            and len(x) == 0
            or hasattr(x, "__iter__")
            and len(list(iter(x))) == 0
        )

    @classmethod
    def is_probable_null(cls, x: Any) -> bool:
        """
        Returns True iff either:
            - x is None
            - np.is_nan(x)
            - x is something with 0 length
            - x is iterable and has 0 elements (will call ``__iter__``)
            - a str(x) is 'nan', 'n/a', 'null', or 'none'; case-insensitive

        Things that are **NOT** probable nulls:
            - "na"
            - 0
            - [None]

        Raises:
            TypeError If ``x`` is an Iterator. Calling this would empty the iterator, which is dangerous.
        """
        return cls.is_empty(x) or str(x).lower() in ["nan", "n/a", "null", "none"]

    @classmethod
    def unique(cls, sequence: Iterable[T]) -> Sequence[T]:
        """
        Returns the unique items in `sequence`, in the order they appear in the iteration.

        Args:
            sequence: Any once-iterable sequence

        Returns:
            An ordered List of unique elements
        """
        seen = set()
        return [x for x in sequence if not (x in seen or seen.add(x))]

    @classmethod
    def first(cls, collection: Iterable[Any], attr: Optional[str] = None) -> Optional[Any]:
        """
        Gets the first element.

        WARNING: Tries to call ``next(x)``, progressing iterators.

        Args:
            collection: Any iterable
            attr: The name of the attribute that might be defined on the elements,
                or None to indicate the elements themselves should be used

        Returns:
            - The attribute of the first element if ``attr`` is defined on an element
            - None if the the sequence is empty
            - None if the sequence has no attribute ``attr``
        """
        try:
            # note: calling iter on an iterator creates a view only
            x = next(iter(collection))
            return x if attr is None else cls.look(x, attr)
        except StopIteration:
            return None

    @classmethod
    def iter_rowcol(cls, n_rows: int, n_cols: int) -> Generator[Tuple[int, int], None, None]:
        """
        An iterator over (row column) pairs for a row-first traversal of a grid with ``n_cols`` columns.

        Example:
            >>> it = CommonTools.iter_rowcol(5, 3)
            >>> [next(it) for _ in range(5)]  # [(0,0),(0,1),(0,2),(1,0),(1,1)]
        """
        for i in range(n_rows * n_cols):
            yield i // n_cols, i % n_cols

    @classmethod
    def multidict(
        cls,
        sequence: Iterable[Z],
        key_attr: Union[str, Iterable[str], Callable[[Y], Z]],
        skip_none: bool = False,
    ) -> Mapping[Y, Sequence[Z]]:
        """
        Builds a mapping from keys to multiple values.
        Builds a mapping of some attribute in ``sequence`` to
        the containing elements of ``sequence``.

        Args:
            sequence: Any iterable
            key_attr: Usually string like 'attr1.attr2'; see `look`
            skip_none: If None, raises a `KeyError` if the key is missing for any item; otherwise, skips it
        """
        dct = defaultdict(lambda: [])
        for item in sequence:
            v = CommonTools.look(item, key_attr)
            if not skip_none and v is None:
                raise KeyError(f"No {key_attr} in {item}")
            if v is not None:
                dct[v].append(item)
        return dct

    @classmethod
    def mem_size(cls, obj) -> str:
        """
        Returns the size of the object in memory as a human-readable string.

        Args:
            obj: Any Python object

        Returns:
            A human-readable size with units
        """
        return nicesize(sys.getsizeof(obj))

    @classmethod
    def devnull(cls):
        """
        Yields a 'writer' that does nothing.

        Example:
            >>> with CommonTools.devnull() as devnull:
            >>>     devnull.write('hello')
        """
        yield DevNull()

    @classmethod
    def parse_bool(cls, s: str) -> bool:
        """
        Parses a 'true'/'false' string to a bool, ignoring case.

        Raises:
            ValueError: If neither true nor false
        """
        if isinstance(s, bool):
            return s
        if s.lower() == "false":
            return False
        if s.lower() == "true":
            return True
        raise ValueError(f"{s} is not true/false")

    @classmethod
    def parse_bool_flex(cls, s: str) -> bool:
        """
        Parses a 'true'/'false'/'yes'/'no'/... string to a bool, ignoring case.

        Allowed:
            - "true", "t", "yes", "y", "1", "+"
            - "false", "f", "no", "n", "0", "-"

        Raises:
            ValueError: If neither true nor false
        """
        mp = {
            **{v: True for v in {"true", "t", "yes", "y", "1", "+"}},
            **{v: False for v in {"false", "f", "no", "n", "0", "-"}},
        }
        v = mp.get(s.lower())
        if v is None:
            raise ValueError(f"{s.lower()} is not in {','.join(mp.keys())}")
        return v


__all__ = ["CommonTools"]



import random
import time
from typing import Callable, Mapping, Optional
from urllib import request


def download_urllib(req: request.Request) -> bytes:
    with request.urlopen(req) as q:
        return q.read()


class QueryExecutor:
    """
    A synchronous GET/POST query executor that limits the rate of requests.
    """

    def __init__(
        self,
        sec_delay_min: float = 0.25,
        sec_delay_max: float = 0.25,
        encoding: Optional[str] = "utf-8",
        querier: Optional[Callable[[request.Request], bytes]] = None,
    ):
        self._min = sec_delay_min
        self._max = sec_delay_max
        self._rand = random.Random()  # nosec
        self._encoding = encoding
        self._next_at = 0
        self._querier = download_urllib if querier is None else querier

    def __call__(
        self,
        url: str,
        method: str = "get",
        encoding: Optional[str] = "-1",
        headers: Optional[Mapping[str, str]] = None,
        errors: str = "ignore",
    ) -> str:
        headers = {} if headers is None else headers
        encoding = self._encoding if encoding == "-1" else encoding
        now = time.monotonic()
        if now < self._next_at:
            time.sleep(self._next_at - now)
        req = request.Request(url=url, method=method, headers=headers)
        content = self._querier(req)
        if encoding is None:
            data = content.decode(errors=errors)
        else:
            data = content.decode(encoding=encoding, errors=errors)
        self._next_at = time.monotonic() + self._rand.uniform(self._min, self._max)
        return data


__all__ = ["QueryExecutor"]



import abc
from collections.abc import Iterator as _Iterator
from typing import Any, Generic, Iterable, TypeVar

import numpy as np

T = TypeVar("T")
IX = TypeVar("IX")


# noinspection PyAbstractClass
class SizedIterator(_Iterator, metaclass=abc.ABCMeta):
    """
    An iterator with size and progress.
    """

    def position(self) -> int:
        raise NotImplementedError()

    def total(self) -> int:
        raise NotImplementedError()

    def remaining(self) -> int:
        return self.total() - self.position()

    def has_next(self) -> bool:
        return self.position() < self.total()

    def __len__(self) -> int:
        return self.total()

    def __repr__(self):
        return "{}({} items)".format(self.__class__, self.total())

    def __str__(self):
        return repr(self)


class SeqIterator(SizedIterator, Generic[IX]):
    """
    A concrete SizedIterator backed by a list.
    """

    def __init__(self, it: Iterable[Any]):
        self.seq, self.__i, self.current = list(it), 0, None

    def reset(self) -> None:
        self.__i, self.current = 0, None

    def peek(self) -> None:
        return self.seq[self.__i]

    def position(self) -> int:
        return self.__i

    def total(self) -> int:
        return len(self.seq)

    def __next__(self) -> T:
        try:
            self.current = self.seq[self.__i]
        except IndexError:
            raise StopIteration(f"Size is {len(self)}")
        self.__i += 1
        return self.current


class TieredIterator(SeqIterator):
    """
    A SizedIterator that iterates over every tuples of combination from multiple sequences.

    Example:
        >>> it = TieredIterator([[1, 2, 3], [5, 6]])
        >>> list(it)
        [(1,5), (1,6), (2,5), (2,6), (3,5), (3,6)]
    """

    # noinspection PyMissingConstructor
    def __init__(self, sequence):
        self.seqs = list([SeqIterator(s) for s in reversed(sequence)])
        self.__total = 0 if len(self.seqs) == 0 else int(np.product([i.total() for i in self.seqs]))
        self.__i = 0

    def position(self) -> int:
        return self.__i

    def total(self) -> int:
        return self.__total

    def __next__(self):
        if not self.has_next():
            raise StopIteration(f"Length is {self.total()}")
        t = tuple((seq.peek() for seq in reversed(self.seqs)))
        self.__set(0)
        self.__i += 1
        return t

    def __set(self, i: int):
        seq = self.seqs[i]
        if seq.remaining() > 1:
            next(seq)
        else:
            seq.reset()
            if (
                i < len(self.seqs) - 1
            ):  # to avoid index error for last case, after which self.has_next() is False
                self.__set(i + 1)  # recurse

    def __repr__(self):
        return "{}({} items)".format(self.__class__.__name__, [it.total() for it in self.seqs])

    def __str__(self):
        return repr(self)


__all__ = ["SizedIterator", "SeqIterator", "TieredIterator"]



from __future__ import annotations

import json
import pickle
from copy import copy
from datetime import date, datetime
from pathlib import Path, PurePath
from typing import Any, ByteString, Callable, Mapping, Optional, Sequence
from typing import Tuple as Tup
from typing import Type, TypeVar, Union

import orjson
import toml

PICKLE_PROTOCOL = 5
T = TypeVar("T")


def _json_encode_default(obj: Any) -> Any:
    if isinstance(obj, NestedDotDict):
        # noinspection PyProtectedMember
        return dict(obj._x)


class NestedDotDict(Mapping):
    """
    A thin wrapper around a nested dict to make getting values easier.
    This was designed as a wrapper for TOML, but it works more generally too.

    Keys must be strings that do not contain a dot (.).
    A dot is reserved for splitting values to traverse the tree.
    For example, ``dotdict["pet.species.name"]``.

    """

    def to_json(self, indent: bool = False) -> str:
        kwargs = dict(option=orjson.OPT_INDENT_2) if indent else {}
        encoded = orjson.dumps(dict(self), default=_json_encode_default, **kwargs)
        return encoded.decode(encoding="utf8")

    @classmethod
    def read_toml(cls, path: Union[PurePath, str]) -> NestedDotDict:
        return NestedDotDict(toml.loads(Path(path).read_text(encoding="utf8")))

    @classmethod
    def read_json(cls, path: Union[PurePath, str]) -> NestedDotDict:
        """
        Reads JSON from a file, into a NestedDotDict.
        If the JSON data is a list type, converts into a dict with the key ``data``.
        """
        data = orjson.loads(Path(path).read_text(encoding="utf8"))
        if isinstance(data, list):
            data = {"data": data}
        return cls(data)

    @classmethod
    def read_pickle(cls, path: Union[PurePath, str]) -> NestedDotDict:
        """

        Note that this function has potential security concerns.
        This is because it relies on the pickle module.
        """
        data = Path(path).read_bytes()
        data = pickle.loads(data)  # nosec
        return cls(data)

    @classmethod
    def parse_toml(cls, data: str) -> NestedDotDict:
        return cls(toml.loads(data))

    @classmethod
    def parse_json(cls, data: str) -> NestedDotDict:
        """
        Parses JSON from a string, into a NestedDotDict.
        If the JSON data is a list type, converts into a dict with the key ``data``.
        """
        data = json.loads(data)
        if isinstance(data, list):
            data = {"data": data}
        return cls(data)

    @classmethod
    def parse_pickle(cls, data: ByteString) -> NestedDotDict:
        if not isinstance(data, bytes):
            data = bytes(data)
        return NestedDotDict(pickle.loads(data))

    def __init__(self, x: Mapping[str, Any]) -> None:
        """
        Constructor.

        Raises:
            ValueError: If a key (in this dict or a sub-dict) is not a str or contains a dot
        """
        if not (hasattr(x, "items") and hasattr(x, "keys") and hasattr(x, "values")):
            raise TypeError(f"Type {type(x)} for value {x} appears not to be dict-like")
        bad = [k for k in x if not isinstance(k, str)]
        if len(bad) > 0:
            raise ValueError(f"Keys were not strings for these values: {bad}")
        bad = [k for k in x if "." in k]
        if len(bad) > 0:
            raise ValueError(f"Keys contained dots (.) for these values: {bad}")
        self._x = x
        # Let's make sure this constructor gets called on subdicts:
        self.leaves()

    def write_json(self, path: Union[PurePath, str], indent: bool = False) -> None:
        kwargs = dict(option=orjson.OPT_INDENT_2) if indent else {}
        Path(path).write_bytes(orjson.dumps(dict(self), default=_json_encode_default, **kwargs))

    def write_pickle(self, path: Union[PurePath, str]) -> None:
        Path(path).write_bytes(pickle.dumps(self._x, protocol=PICKLE_PROTOCOL))

    def leaves(self) -> Mapping[str, Any]:
        """
        Gets the leaves in this tree.

        Returns:
            A dict mapping dot-joined keys to their values
        """
        mp = {}
        for key, value in self._x.items():
            if len(key) == 0:
                raise AssertionError(f"Key is empty (value=${value})")
            if isinstance(value, dict):
                mp.update({key + "." + k: v for k, v in NestedDotDict(value).leaves().items()})
            else:
                mp[key] = value
        return mp

    def sub(self, items: str) -> NestedDotDict:
        return NestedDotDict(self[items])

    def exactly(self, items: str, astype: Type[T]) -> T:
        """
        Gets the key ``items`` from the dict if it has type ``astype``.
        Calling ``dotdict.exactly(k, t) is equivalent to calling ``t(dotdict[k])``,
        but a raised ``TypeError`` will note the key, making this a useful shorthand for the above within a try-except.

        Args:
            items: The key hierarchy, with a dot (.) as a separator
            astype: The type, which will be checked using ``isinstance``

        Returns:
            The value in the required type

        Raises:
            TypeError: If not ``isinstance(value, astype)``
        """
        z = self[items]
        if not isinstance(z, astype):
            raise TypeError(f"Value {z} from {items} is a {type(z)}, not {astype}")
        return z

    def get_as(
        self, items: str, astype: Callable[[Any], T], default: Optional[T] = None
    ) -> Optional[T]:
        """
        Gets the value of an *optional* key, or ``default`` if it doesn't exist.
        Also see ``req_as``.

        Args:
            items: The key hierarchy, with a dot (.) as a separator.
                   Ex: ``animal.species.name``.
            astype: Any function that converts the found value to type ``T``.
                    Can be a ``Type``, such as ``int``.
                    Despite the annotated type, this function only needs to accept the actual value of the key
                    as input, not ``Any``.
            default: Return this value if the key is not found (at any level)

        Returns:
            The value of found key in this dot-dict, or ``default``.

        Raises:
            ValueError: Likely exception raised if calling ``astype`` fails
        """
        x = self.get(items)
        if x is None:
            return default
        if astype is date:
            return self._to_date(x)
        if astype is datetime:
            return self._to_datetime(x)
        return astype(x)

    def req_as(self, items: str, astype: Optional[Callable[[Any], T]]) -> T:
        """
        Gets the value of a *required* key.
        Also see ``get_as`` and ``exactly``.

        Args:
            items: The key hierarchy, with a dot (.) as a separator.
                   Ex: ``animal.species.name``.
            astype: Any function that converts the found value to type ``T``.
                    Can be a ``Type``, such as ``int``.
                    Despite the annotated type, this function only needs to accept the actual value of the key
                    as input, not ``Any``.

        Returns:
            The value of found key in this dot-dict.

        Raises:
            KeyError: If the key is not found (at any level).
            ValueError: Likely exception raised if calling ``astype`` fails
        """
        x = self[items]
        return astype(x)

    def get_list_as(
        self, items: str, astype: Callable[[Any], T], default: Optional[Sequence[T]] = None
    ) -> Optional[Sequence[T]]:
        """
        Gets list values from an *optional* key.
        Note that ``astype`` here converts elements *within* the list, not the whole list.
        Also see ``req_list_as``.

        Args:
            items: The key hierarchy, with a dot (.) as a separator. Ex: ``animal.species.name``.
            astype: Any function that converts the found value to type ``T``. Ex: ``int``.
            default: Return this value if the key wasn't found

        Returns:
            ``[astype(v) for v in self[items]]``, or ``default`` if ``items`` was not found.

        Raises:
            ValueError: Likely exception raised if calling ``astype`` fails
            TypeError: If the found value is not a (non-``str``) ``Sequence``
        """
        x = self.get(items)
        if x is None:
            return default
        if not isinstance(x, Sequence) or isinstance(x, str):
            raise TypeError(f"Value {x} is not a list for lookup {items}")
        return [astype(y) for y in x]

    def req_list_as(self, items: str, astype: Optional[Callable[[Any], T]]) -> Sequence[T]:
        """
        Gets list values from a *required* key.
        Note that ``astype`` here converts elements *within* the list, not the whole list.
        Also see ``get_list_as``.

        Args:
            items: The key hierarchy, with a dot (.) as a separator. Ex: ``animal.species.name``.
            astype: Any function that converts the found value to type ``T``. Ex: ``int``.

        Returns:
            ``[astype(v) for v in self[items]]``

        Raises:
            ValueError: Likely exception raised if calling ``astype`` fails
            TypeError: If the found value is not a (non-``str``) ``Sequence``
            KeyError: If the key was not found (at any level)
        """
        x = self[items]
        if not isinstance(x, Sequence) or isinstance(x, str):
            raise TypeError(f"Value {x} is not a list for lookup {items}")
        return [astype(y) for y in x]

    def get(self, items: str, default: Any = None) -> Any:
        """
        Gets a value from an optional key.
        Also see ``__getitem__``.
        """
        try:
            return self[items]
        except KeyError:
            return default

    def __getitem__(self, items: str) -> Any:
        """
        Gets a value from a required key.
        Analogous to ``dict.__getitem__``, but this can operate on dot-joined strings.

        **NOTE:** The number of keys for which this returns a value can be different from ``len(self)``.

        Example:
            >>> d = NestedDotDict(dict(a=dict(b=1)))
            >>> assert d["a.b"] == 1
        """
        at = self._x
        for item in items.split("."):
            if item not in at:
                raise KeyError(f"{items} not found: {item} does not exist")
            at = at[item]
        return NestedDotDict(at) if isinstance(at, dict) else copy(at)

    def items(self) -> Sequence[Tup[str, Any]]:
        return list(self._x.items())

    def keys(self) -> Sequence[str]:
        return list(self._x.keys())

    def values(self) -> Sequence[Any]:
        return list(self._x.values())

    def pretty_str(self) -> str:
        """
        Pretty-prints the leaves of this dict using ``json.dumps``.

        Returns:
            A multi-lined string
        """
        return json.dumps(
            self.leaves(),
            sort_keys=True,
            indent=4,
            ensure_ascii=False,
        )

    def __len__(self) -> int:
        """
        Returns the number of values in this dict.
        Does **NOT** include nested values.
        """
        return len(self._x)

    def __iter__(self):
        """
        Iterates over values in this dict.
        Does **NOT** include nested items.
        """
        return iter(self._x)

    def __repr__(self):
        return repr(self._x)

    def __str__(self):
        return str(self._x)

    def __eq__(self, other):
        return str(self) == str(other)

    def _to_date(self, s):
        if isinstance(s, date):
            return s
        elif isinstance(s, str):
            # This is MUCH faster than tomlkit's
            return date.fromisoformat(s)
        else:
            raise TypeError(f"Invalid type ${type(s)} for {s}")

    def _to_datetime(self, s):
        if isinstance(s, datetime):
            return s
        elif isinstance(s, str):
            # This is MUCH faster than tomlkit's
            if s.count(":") < 2:
                raise ValueError(f"Datetime {s} does not contain hours, minutes, and seconds")
            return datetime.fromisoformat(s.upper().replace("Z", "+00:00"))
        else:
            raise TypeError(f"Invalid type ${type(s)} for {s}")


__all__ = ["NestedDotDict"]



import itertools
import logging
import multiprocessing
import time
import warnings
from datetime import datetime
from typing import (
    Any,
    Callable,
    Collection,
    Generator,
    Iterable,
    Iterator,
    Optional,
    TypeVar,
    Union,
)

from pocketutils.tools.base_tools import BaseTools
from pocketutils.tools.unit_tools import UnitTools

logger = logging.getLogger("pocketutils")
T = TypeVar("T")
K = TypeVar("K", covariant=True)
V = TypeVar("V", contravariant=True)


class LoopTools(BaseTools):
    @classmethod
    def loop(
        cls,
        things: Iterable[T],
        *,
        log: Union[None, str, Callable[[str], Any]] = None,
        every_i: int = 10,
        n_total: Optional[int] = None,
    ) -> Generator[T, None, None]:
        """
        Loops over elements while logging time taken, % processed, etc.

        Args:
            things: Items to process; if sized or ``n_total`` is passed,
                    will log estimates and % complete
            log: Write by calling this function.
                 The returned value is ignored.
                 if ``str``, gets a function from python logging
                 (via :attr:`pocketutils.tools.loop_tools.logger`);
                 see :meth:`get_log_function`.
                 If ``None``, does not log.
            every_i: Log after every ``every_i`` items processed
            n_total: Provide if ``things`` is not sized yet the total number
                     of items is known
        """
        log = cls.get_log_function(log)
        if hasattr(things, "__len__") or n_total is not None:
            # noinspection PyTypeChecker
            yield from cls._loop_timing(things, log, every_i, n_total)
        else:
            yield from cls._loop_logging(things, log, every_i)

    @classmethod
    def parallel(
        cls,
        items: Collection[K],
        function: Callable[[K], V],
        *,
        to=print,
        n_cores: int = 2,
        poll_sec: float = 0.4,
    ) -> None:
        """
        Process items with multiprocessing and a rotating cursor with % complete.

        Args:
            items: Items to process; must have ``__len__`` defined
            function: Called per item
            to: Write (log) to this function; ``end="\r"`` is passed for the cursor
            n_cores: The number pool cores
            poll_sec: Check for new every ``poll_sec`` seconds
        """
        t0 = time.monotonic()
        if to is not None:
            to(f"Using {n_cores} cores...")
        with multiprocessing.Pool(n_cores) as pool:
            queue = multiprocessing.Manager().Queue()
            result = pool.starmap_async(function, items)
            cycler = itertools.cycle(r"\|/―")
            while not result.ready():
                percent = queue.qsize() / len(items)
                if to is not None:
                    to(f"% complete: {percent:.0%} {next(cycler)}", end="\r")
                time.sleep(poll_sec)
            got = result.get()
        delta = UnitTools.delta_time_to_str(time.monotonic() - t0)
        if to is not None:
            to(f"\nProcessed {len(got)} items in {delta}")

    @classmethod
    def _loop_logging(
        cls,
        things: Iterable[T],
        log: Union[None, str, Callable[[str], None]] = None,
        every_i: int = 10,
    ) -> Iterator[T]:
        log = cls.get_log_function(log)
        initial_start_time = time.monotonic()
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        log(f"Started processing at {now}.")
        i = 0
        for i, thing in enumerate(things):
            t0 = time.monotonic()
            yield thing
            t1 = time.monotonic()
            if i % every_i == 0 and i > 0:
                elapsed_s = UnitTools.delta_time_to_str(t1 - t0)
                log(f"Processed next {every_i} in {elapsed_s}")
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        delta = UnitTools.delta_time_to_str(time.monotonic() - initial_start_time)
        log(f"Processed {i}/{i} in {delta}. Done at {now}.")

    @classmethod
    def _loop_timing(
        cls,
        things: Collection[Any],
        log: Union[None, str, Callable[[str], None]] = None,
        every_i: int = 10,
        n_total: Optional[int] = None,
    ):
        log = cls.get_log_function(log)
        n_total = len(things) if n_total is None else n_total
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        log(f"Started processing {n_total} items at {now}.")
        t0 = time.monotonic()
        initial_start_time = time.monotonic()
        for i, thing in enumerate(things):
            yield thing
            t1 = time.monotonic()
            if i % every_i == 0 and i < n_total - 1:
                estimate = (t1 - initial_start_time) / (i + 1) * (n_total - i - 1)
                elapsed_s = UnitTools.delta_time_to_str(t1 - t0)
                estimate_s = UnitTools.delta_time_to_str(estimate)
                log(f"Processed {i+1}/{n_total} in {elapsed_s}. Estimated {estimate_s} left.")
        now = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        delta = UnitTools.delta_time_to_str(time.monotonic() - initial_start_time)
        log(f"Processed {n_total}/{n_total} in {delta}. Done at {now}.")


__all__ = ["LoopTools"]



import subprocess  # nosec
import warnings
from dataclasses import dataclass

import regex

from pocketutils.core import PathLike
from pocketutils.core.exceptions import ParsingError
from pocketutils.tools.base_tools import BaseTools


@dataclass(frozen=True)
class GitDescription:
    """
    Data collected from running `git describe --long --dirty --broken --abbrev=40 --tags`
    """

    text: str
    tag: str
    commits: str
    hash: str
    is_dirty: bool
    is_broken: bool

    def __repr__(self):
        return self.__class__.__name__ + "(" + self.text + ")"

    def __str__(self):
        return repr(self)


class ProgramTools(BaseTools):
    """

    Security concerns
    -----------------

    Please note that these tools execute external code
    through the ``subprocess`` module.
    These calls are additionally made on partial executable paths,
    such as ``git`` rather than ``/usr/bin/git``.
    This is an additional security consideration.
    """

    @classmethod
    def commit_hash(cls, git_repo_dir: str = ".") -> str:
        """
        Gets the hex of the most recent Git commit hash in git_repo_dir.
        """
        warnings.warn(
            "commit_hash will be removed; use git_description instead", DeprecationWarning
        )
        return cls.git_description(git_repo_dir).hash

    @classmethod
    def git_description(cls, git_repo_dir: PathLike = ".") -> GitDescription:
        """
        Runs ``git describe`` and parses the output.

        Args:
            git_repo_dir: Path to the repository

        Returns:
            A ``GitDescription`` instance, with fields text, tag, commits, hash, is_dirty, and is_broken

        Raises:
            CalledProcessError:
        """
        cmd_args = dict(
            cwd=str(git_repo_dir),
            capture_output=True,
            check=True,
            text=True,
            encoding="utf8",
        )
        cmd = "git describe --long --dirty --broken --abbrev=40 --tags".split(" ")
        # ignoring bandit security warning because we explain the security concerns
        # in the class docstring
        x = subprocess.run(cmd, **cmd_args)  # nosec
        return cls._parse(x.stdout.decode(encoding="utf8").strip())

    @classmethod
    def _parse(cls, text: str):
        pat = regex.compile(r"([\d.]+)-(\d+)-g([0-9a-h]{40})(?:-([a-z]+))?", flags=regex.V1)
        # ex: 1.8.6-43-g0ceb89d3a954da84070858319f177abe3869752b-dirty
        m = pat.fullmatch(text)
        if m is None:
            raise ParsingError(f"Bad git describe string {text}")
        # noinspection PyArgumentList
        return GitDescription(
            text,
            m.group(1),
            m.group(2),
            m.group(3),
            m.group(4) == "dirty",
            m.group(4) == "broken",
        )


__all__ = ["GitDescription", "ProgramTools"]



import enum
import gzip
import logging
import shutil
import zipfile
from datetime import datetime
from pathlib import Path
from typing import Mapping, Optional
from urllib import request

from pocketutils.core import PathLike

logger = logging.getLogger("pocketutils")


class WebResource:
    """
    Useful for extracting files from ZIP and GZIPing them.
    """

    def __init__(self, url: str, archive_member: Optional[str], local_path: PathLike):
        self._url = url
        self._archive_member = archive_member
        self._local_path = Path(local_path)

    def download(self, redownload: bool = False):
        now = datetime.now()
        to_path = Path(self._local_path)
        if not to_path.exists() or redownload:
            extracted, dled = None, None
            try:
                logger.info(f"Downloading {self._url}...")
                tmp = str(to_path) + now.strftime("%Y%m%d-%H%M%S-%f") + ".tmp"
                dled, response = request.urlretrieve(self._url, tmp)
                dled = Path(dled)
                if self._archive_member is not None:
                    with zipfile.ZipFile(dled, "r") as zfile:
                        extracted = Path(zfile.extract(self._archive_member))
                else:
                    extracted = dled
                if to_path.suffix == ".gz" and not self.__is_gzip(extracted):
                    with extracted.open("rb") as f_in:
                        with gzip.open(to_path, "wb") as f_out:
                            shutil.copyfileobj(f_in, f_out)
                else:
                    shutil.move(extracted, to_path)
                self._info_path.write_text(
                    "url="
                    + self._url
                    + "\n"
                    + "datetime_downloaded="
                    + now.isoformat()
                    + "\n"
                    + "response="
                    + str(response).replace("\n", " |")
                    + "\n",
                    encoding="utf-8",
                )
            finally:
                if extracted is not None and extracted.exists():
                    extracted.unlink()
                if dled is not None and dled.exists():
                    dled.unlink()

    def datetime_downloaded(self) -> datetime:
        return datetime.fromisoformat(self.metadata()["datetime_downloaded"])

    def metadata(self) -> Mapping[str, str]:
        return {
            line[: line.index("=")].strip(): line[line.index("=") + 1 :].strip()
            for line in self._info_path.read_text(encoding="utf8").splitlines()
        }

    @property
    def _info_path(self) -> Path:
        return self._local_path.with_suffix(self._local_path.suffix + ".info")

    def exists(self) -> bool:
        return self._local_path.exists()

    @property
    def path(self) -> Path:
        return self._local_path

    def delete(self) -> None:
        self.path.unlink(missing_ok=True)

    def __is_gzip(self, path):
        try:
            with gzip.open(path, "rb") as f:
                f.read(20)  # 10-byte header
        except OSError as e:
            if "Not a gzipped file" in str(e):
                return False
            else:
                raise e
        return True


__all__ = ["WebResource"]



"""
The motivation here is simply that Python lacks some standard exceptions that I consider important.
Projects can/should subclass from these in addition to the normal Python ones.
"""
from __future__ import annotations

import logging
import os
from copy import copy
from functools import wraps
from pathlib import Path

# noinspection PyUnresolvedReferences
from subprocess import CalledProcessError
from typing import Any, Collection, Union

logger = logging.getLogger("pocketutils")
KeyLike = Any
PathLike = Union[Path, str, os.PathLike]
_DW = DeprecationWarning
_PDW = PendingDeprecationWarning


###################################################################################################
#                                                                                                 #
#                                             utilities                                           #
#                                                                                                 #
###################################################################################################


class ErrorUtils:
    """Utilities for creating and modifying errors."""

    @staticmethod
    def args(**names):
        """
        Decorator.
        Add a __init__ that calls the superclass and takes any argument in names.
        The __init__ will set: (self.name=value if name is passed else None) for all name in names.

        Args:
            names: A map from str to Type
        """
        if any([s == "info" or s.startswith("__") and s.endswith("__") for s in names]):
            raise AssertionError(f"Failed on {names}")

        @wraps(names)
        def dec(cls):
            def _doc(doc: str) -> str:
                if doc is None:
                    doc = ""
                elif not doc.endswith("."):
                    doc += "."
                return (
                    doc
                    + "\n"
                    + "Supports attributes:\n"
                    + "\n".join(
                        [
                            "    • "
                            + name
                            + ": "
                            + str(dtype).replace("<class '", "").replace("'>", "")
                            for name, dtype in names.items()
                        ]
                    )
                )

            def _init(self, *args, **kwargs) -> None:
                kwargs = copy(kwargs)
                # when we call super(), we need to know which class we're on and which should be called next
                # note that self will always be the first class
                thisclass = kwargs.pop("__thisclass") if "__thisclass" in kwargs else self.__class__
                for name, reqtype in names.items():
                    if name in kwargs:
                        value = kwargs.pop(name)
                        # TODO in the future, we can warn when type-checks fail; impossible in current Python
                        setattr(self, name, value)
                    else:
                        setattr(self, name, None)
                    nextclass = thisclass.__mro__[1]
                    # noinspection PyArgumentList
                    # be careful! use `thisclass` as an argument (not self)
                    super(thisclass, self).__init__(*args, __thisclass=nextclass, **kwargs)

            cls.__init__ = _init
            cls.__doc__ = _doc(cls.__doc__)
            return cls

        return dec


class _FnUtils:
    """
    Abstract class for warnings and errors that can contain extra attributes.
    """

    def eq(self, other) -> bool:
        return (
            type(self) == type(other)
            and str(self) == str(other)
            and all([getattr(self, name) == getattr(other, name) for name in self.__dict__])
        )

    def info(self) -> str:
        return (
            self.__class__.__name__
            + ":"
            + str(self)
            + "("
            + ",".join([k + "=" + str(v) for k, v in sorted(self.__dict__.items())])
            + ")"
        )


###################################################################################################
#                                                                                                 #
#                                           warning types                                         #
#                                                                                                 #
###################################################################################################


# NOTE: line 2nd col at 50 chars, no matter what
# This indentation makes it easy to see new definitions
# If not possible, adjust only for that line
# Don't adjust existing lines from 50 to avoid git tracking


class XException(Exception):
    """Abstract exception whose subclasses can define extra attributes using ErrorTools.args."""

    def __init__(self, *args, **kwargs):
        pass

    info = _FnUtils.info
    __eq__ = _FnUtils.eq


class XWarning(UserWarning):
    """Abstract user warning whose subclasses can define extra attributes using ErrorTools.args."""

    def __init__(self, *args, **kwargs):
        pass

    info = _FnUtils.info
    __eq__ = _FnUtils.eq


class _CodeWarning(XWarning):
    """A warning related to code quality."""


class ObsoleteWarning(_CodeWarning, _PDW):
    """The code being called is obsolete and/or may be deprecated in the future."""


class DeprecatedWarning(_CodeWarning, _DW):
    """The code being called is deprecated."""


class ImmatureWarning(_CodeWarning):
    """The code being called is unstable or immature."""


class _OpWarning(XWarning):
    """A warning related to problematic requests."""


class DangerousRequestWarning(_OpWarning):
    """A warning about an operation that is dangerous but is proceeding regardless."""


class StrangeRequestWarning(_OpWarning):
    """A warning about a potential result being wrong because weird arguments were passed."""


class IgnoringRequestWarning(_OpWarning):
    """A request or passed argument was ignored."""


class _MissingWarning(XWarning):
    """A request related to something missing."""


@ErrorUtils.args(package=str)
class ImportFailedWarning(_MissingWarning):
    """Could not import a recommended package."""


@ErrorUtils.args(key=KeyLike)
class ConfigWarning(_MissingWarning):
    """Did not find a recommended config entry, etc."""


class AlgorithmWarning(XWarning):
    """A warning about a potential result being wrong."""


@ErrorUtils.args(data_key=KeyLike)
class DataWarning(XWarning):
    """External data is suspicious / might contain an error."""


###################################################################################################
#                                                                                                 #
#                                            error types                                          #
#                                                                                                 #
###################################################################################################


class Error(XException):
    """
    Abstract exception that could reasonably be recovered from.
    Subclass names should end with 'Error'.
    """


class NaturalExpectedError(Error):
    """Non-specific exception to short-circuit behavior but meaning 'all ok'."""


@ErrorUtils.args(name=str)
class CodeIncompleteError(Error, NotImplementedError):
    """The code is not finished!"""


class IllegalStateError(Error, AssertionError):
    """
    An assertion failed marking invalid state, potentially recoverable.
    See Also:
         :class:`OpStateError`
    """


class UnsupportedOpError(Error):
    """
    Used as a replacement for NotImplementedError, where the method *should not* be implemented.
    This also differs from ``NotImplemented`` in that is permanent
    and that it should not (usually) be raised in special methods like ``__lt__``.
    """


class NotConstructableError(UnsupportedOpError):
    """
    Constructing this object is not supported.
    It's probably a singleton, static utils, or has factory methods.
    """


@ErrorUtils.args(class_name=str)
class ImmutableError(UnsupportedOpError):
    """Tried to mutate an immutable object."""


class OpStateError(UnsupportedOpError):
    """
    The operation cannot be performed on an object in this state.
    See Also:
         :class:`IllegalStateError`
    """


@ErrorUtils.args(n=int)
class MultipleMatchesError(Error):
    """
    Multiple records match when only 1 was expected.
    Also applies if 0 or records are permitted.
    Has argument ``n`` containing the number.
    """


class UserError(Error):
    """An error caused by input from a user of an application."""


@ErrorUtils.args(cmd=str)
class BadCommandError(UserError):
    """The user passed an invalid command, such as at the command-line."""


class ConfigError(UserError):
    """Error while interpreting a config file from the user."""


class MissingConfigKeyError(UserError):
    """Missing a required key."""


class RefusingRequestError(UserError):
    """The user requested an operation that is understood but is being refused."""


@ErrorUtils.args(key=KeyLike)
class ResourceError(Error):
    """A problem finding or loading a resource (file, network connection, hardware, etc.)."""


class LockedError(ResourceError):
    """A resource was found but is locked (ex a hardware component in use)."""


class MissingEnvVarError(ResourceError):
    """Missing a required environment variable."""


@ErrorUtils.args(expected=str, actual=str)
class HashValidationFailedError(ResourceError):
    """A checksum did not validate."""


HashValidationError = HashValidationFailedError


class IncompatibleDataError(ResourceError):
    """An operation cannot be applied to data with this property value."""


class MismatchedDataError(ResourceError):
    """A property of the data is different between multiple elements."""


class MissingResourceError(ResourceError):
    """Could not find a resource (general case)."""


class LookupFailedError(MissingResourceError):
    """Could not find a resource by name."""


class RequestError(Error):
    """A generic error related to arguments."""


@ErrorUtils.args(key=KeyLike)
class RefusingError(RequestError):
    """
    General-case refusal to run an operation, not clearly caused by an application user.

    See Also:
        RefusingRequestError
    """


class AmbiguousRequestError(RequestError):
    """Insufficient information was passed to resolve the operation."""


class ContradictoryRequestError(RequestError):
    """Contradictory information was passed."""


@ErrorUtils.args(key=KeyLike)
class ReservedError(RequestError):
    """A key is reserved by the code and cannot be used."""


@ErrorUtils.args(key=KeyLike)
class AlreadyUsedError(RequestError):
    """A key was specified twice."""


class _SizeError(Error):
    """Too small or large for the operation."""


@ErrorUtils.args(length=int, minimum=int, maximum=int)
class LengthError(_SizeError):
    """The length is too large or too small."""


@ErrorUtils.args(lengths=Collection[int])
class LengthMismatchError(_SizeError):
    """The objects (2 or more) have different lengths."""


@ErrorUtils.args(expected=int, actual=int)
class WrongDimensionError(_SizeError):
    """The object has the wrong number of dimensions (ex matrix instead of vector)"""


class EmptyCollectionError(_SizeError):
    """The object has no elements."""


@ErrorUtils.args(expected=str, actual=str)
class XTypeError(Error, TypeError):
    """A TypeError containing the expected and actual types."""


@ErrorUtils.args(value=KeyLike)
class XValueError(Error, ValueError):
    """A ValueError containing the value."""


@ErrorUtils.args(pattern=str)
class StringPatternError(XValueError):
    """A string did not match a required regular expression."""


@ErrorUtils.args(minimum=int, maximum=int)
class OutOfRangeError(XValueError):
    """A numerical value is outside a required range."""


class ZeroDistanceError(XValueError):
    """Two (or more) values are the same by distance."""


class NullValueError(XValueError):
    """A value of None, NaN, or similar was given."""


@ErrorUtils.args(value=KeyLike)
class _NumericError(Error):
    """An error with a (simple) numeric operation."""


class NumericConversionError(_NumericError):
    """Could not convert one numeric type to another."""


class InexactRoundError(_NumericError):
    """A floating-point number could not be cast to an integer."""


@ErrorUtils.args(key=KeyLike)
class XKeyError(Error, KeyError):
    """KeyError that contains the failed key."""


# parsing files or similar resources
@ErrorUtils.args(resource=Any)
class _ParsingLikeError(Error):
    """Failed to parse."""


class ParsingError(Error):
    """Syntax error when parsing."""


@ErrorUtils.args(item=KeyLike)
class UnrecognizedKeyError(_ParsingLikeError):
    """A configuration entry was set, but the code doesn't recognize that name."""


@ErrorUtils.args(item=KeyLike)
class AbstractWrappedError(Error):
    pass


class DataIntegrityError(AbstractWrappedError):
    """Data is missing, incomplete, or invalid. More complex than a missing value."""


class AlgorithmError(AbstractWrappedError):
    """A wrapper for some less meaningful error."""


class ConstructionError(AbstractWrappedError):
    """A wrapper for some less meaningful error."""


class _IoError(Error, IOError):
    """Error related to the filesystem or I/O."""


@ErrorUtils.args(device=KeyLike)
class HardwareError(_IoError):
    """Error related to hardware such as an Arduino."""


class DeviceConnectionError(HardwareError):
    """ "Could not connect to the device."""


class MissingDeviceError(HardwareError):
    """ "Could not find the needed device."""


@ErrorUtils.args(key=KeyLike, value=Any)
class BadWriteError(HardwareError):
    """Could not write to a hardware device (typically to a pin)."""


@ErrorUtils.args(key=KeyLike)
class _LoadSaveError(_IoError):
    """Error loading or saving a file."""


class LoadError(_LoadSaveError):
    """Failed to load a file."""


class SaveError(_LoadSaveError):
    """Failed to save a file."""


class CacheLoadError(_LoadSaveError):
    """Failed to load a file from a cache."""


class CacheSaveError(_LoadSaveError):
    """Failed to save a file to a cache."""


class DownloadError(_LoadSaveError):
    """Failed to download a file."""


class UploadError(_LoadSaveError):
    """Failed to upload a file."""


@ErrorUtils.args(path=PathLike)
class PathError(_IoError):
    """Error involving a path on the filesystem."""


class IllegalPathError(PathError, ValueError):
    """Not a valid path (e.g. not ok on the filesystem)."""


class FileDoesNotExistError(PathError):
    """A file is expected, but the path does not exist."""


class DirDoesNotExistError(PathError):
    """A directory is expected, but the path does not exist."""


class PathExistsError(PathError):
    """The path already exists."""


class XFileExistsError(PathError, FileExistsError):
    """The file already exists."""


class PathIsNotAFileError(PathExistsError):
    """The path already exists and is not a file."""


class PathIsNotADirError(PathExistsError):
    """The path already exists and is not a directory."""



import operator
import sys
from abc import ABC
from datetime import datetime
from typing import Any, Callable, Dict, Generic, Iterator, Mapping, Optional, Sequence, TypeVar

from psutil import virtual_memory

# noinspection PyProtectedMember
from pocketutils.core._internal import nicesize

K = TypeVar("K", covariant=True)
V = TypeVar("V", contravariant=True)


class MemCachePolicy(Generic[K, V]):
    def should_archive(self) -> bool:
        raise NotImplementedError()

    def can_archive(self) -> bool:
        raise NotImplementedError()

    def reindex(self, items: Mapping[K, V]) -> None:
        raise NotImplementedError()

    def items(self) -> Iterator[K]:
        # TODO this is not overloaded!!
        raise NotImplementedError()

    def accessed(self, key: K) -> None:
        pass

    def added(self, key: K, value: V) -> None:
        pass

    def removed(self, key: K) -> None:
        pass


class MemoryLimitingPolicy(MemCachePolicy, ABC):
    def __init__(
        self,
        max_memory_bytes: Optional[int] = None,
        max_fraction_available_bytes: Optional[float] = None,
    ):
        self._max_memory_bytes = max_memory_bytes
        self._max_fraction_available_bytes = max_fraction_available_bytes
        self._total_memory_bytes = 0
        self._usage_bytes = {}  # type: Dict[K, int]
        self._last_accessed = {}  # type: Dict[K, datetime]
        self._created = {}  # type: Dict[K, datetime]

    def can_archive(self) -> bool:
        return len(self._last_accessed) > 0

    def should_archive(self) -> bool:
        return (
            self._max_memory_bytes is not None and self._total_memory_bytes > self._max_memory_bytes
        ) or (
            self._max_fraction_available_bytes is not None
            and self._total_memory_bytes
            > virtual_memory().available * self._max_fraction_available_bytes
        )

    def reindex(self, items: Mapping[K, V]) -> None:
        for key in set(self._last_accessed.keys()) - set(items.keys()):
            if key not in items.keys():
                self.removed(key)
        for key, value in items.items():
            self._usage_bytes[key] = sys.getsizeof(value)
        self._total_memory_bytes = sum(self._usage_bytes.values())

    def accessed(self, key: K) -> None:
        self._last_accessed[key] = datetime.now()

    def added(self, key: K, value: V) -> None:
        now = datetime.now()
        self._created[key] = now
        self._last_accessed[key] = now
        self._usage_bytes[key] = sys.getsizeof(value)
        self._total_memory_bytes += self._usage_bytes[key]

    def removed(self, key: K) -> None:
        self._total_memory_bytes -= self._usage_bytes[key]
        del self._last_accessed[key]
        del self._created[key]
        del self._usage_bytes[key]

    def __str__(self):
        available = virtual_memory().available
        name = self.__class__.__name__
        n_items = len(self._usage_bytes)
        real = nicesize(self._total_memory_bytes)
        cap_raw = "-" if self._max_memory_bytes is None else nicesize(self._max_memory_bytes)
        cap_percent = (
            "-"
            if self._max_fraction_available_bytes is None
            else nicesize(available * self._max_fraction_available_bytes)
        )
        percent = (
            "-"
            if self._max_fraction_available_bytes is None
            else round(
                100 * self._total_memory_bytes / (available * self._max_fraction_available_bytes),
                3,
            )
        )
        return f"{name}(n={n_items}, {real}/{cap_raw}, {real}/{cap_percent}={percent}%"

    def __repr__(self):
        ordered = list(self.items())
        ss = []
        current_day = None
        for k in ordered:
            dt = self._last_accessed[k]
            if current_day is None or current_day.date() != dt.date():
                current_day = dt
                ss.append("#" + current_day.strftime("%Y-%m-%d") + "...")
            nice = nicesize(self._usage_bytes[k])
            access = self._last_accessed[k].strftime("%H:%M:%S")
            ss.append(f"{k}:{nice}@{access}")
        return f"{str(self)}@{hex(id(self))}: [{', '.join(ss)}]"


class MemoryLruPolicy(MemoryLimitingPolicy):
    def items(self) -> Iterator[K]:
        return iter([k for k, v in sorted(self._last_accessed.items(), key=operator.itemgetter(1))])


class MemoryMruPolicy(MemoryLimitingPolicy):
    def items(self) -> Iterator[K]:
        rev = reversed(sorted(self._last_accessed.items(), key=operator.itemgetter(1)))
        return iter([k for k, v in rev])


class MemCache(Generic[K, V]):
    def __init__(
        self, loader: Callable[[K], V], policy: MemCachePolicy, *, log: Callable[[str], Any]
    ):
        self._loader = loader
        self._items = {}  # type: Dict[K, V]
        self._policy = policy
        self._log = lambda _: None if log is None else log

    def __getitem__(self, key: K) -> V:
        self._policy.accessed(key)
        if key in self._items:
            return self._items[key]
        else:
            value = self._loader(key)
            self._log(f"Loaded {key}")
            self._items[key] = value
            self._policy.added(key, value)
            self.archive()
            return value

    def __call__(self, key: K) -> V:
        return self[key]

    def archive(self, at_least: Optional[int] = None) -> Sequence[K]:
        it = self._policy.items()
        archived = []
        while self._policy.can_archive() and (
            at_least is not None and len(archived) < at_least or self._policy.should_archive()
        ):
            key = next(it)
            self._policy.removed(key)
            del self._items[key]
            archived.append(key)
            self._log(f"Archived {len(archived)} items: {archived}")
        return archived

    def clear(self) -> None:
        it = self._policy.items()
        cleared = []
        while self._policy.can_archive():
            key = next(it)
            self._policy.removed(key)
            del self._items[key]
            cleared.append(key)
        self._log(f"Cleared {len(cleared)} items: {cleared}")

    def remove(self, key: K) -> None:
        if key in self:
            self._policy.removed(key)
            del self._items[key]

    def __contains__(self, key: K):
        return key in self._items

    def __delitem__(self, key):
        self.remove(key)

    def __repr__(self):
        return str(self) + "@" + hex(id(self))

    def __str__(self):
        return f"{self.__class__.__name__}({repr(self._policy)})"


__all__ = [
    "MemCachePolicy",
    "MemCache",
    "MemoryLimitingPolicy",
    "MemoryLruPolicy",
    "MemoryMruPolicy",
]



import json
import logging
from datetime import datetime
from functools import total_ordering
from pathlib import Path
from typing import Any, Iterator, Sequence, Set

import regex
import requests

from pocketutils.core import PathLike
from pocketutils.core.exceptions import IllegalStateError

logger = logging.getLogger("pocketutils")


@total_ordering
class Atc:
    def __init__(self, code: str, desc: str, parent):
        self.code = code
        self.desc = desc
        self.parent = parent
        self.level = len(code) if self.code != "/" else 0
        self.children = set()

    def __eq__(self, o):
        return isinstance(o, Atc) and self.code == o.code

    def __hash__(self):
        return hash(self.code)

    def __lt__(self, o):
        return isinstance(o, Atc) and self.code < o.code

    def __repr__(self):
        return "→{:7} – {}".format(self.code, self.desc)

    def __str__(self):
        return repr(self)

    def is_root(self) -> bool:
        return self.parent is None

    def is_leaf(self) -> bool:
        if len(self.children) == 0 and self.level == 7:
            return True
        elif len(self.children) != 0 and self.level != 7:
            return False
        raise IllegalStateError(f"Inconsistent leaf definition for {self}")

    def ancestry(self) -> Sequence[Any]:
        parents = []
        if self.parent is not None:
            for a in self.parent.ancestry():
                parents.append(a)
        parents.append(self)
        return parents

    def descendents(self):
        return {b for b in self.bfs() if b != self}

    def subtree(self):
        return AtcTree(self, {n.code: n for n in self.bfs()})

    def dfs(self):
        for node in self.children:
            yield from node.dfs()
            yield node
        yield self

    def bfs(self):
        yield self
        for node in self.children:
            yield node
            yield from node.dfs()

    def leaves(self):
        for node in self.bfs():
            if node.level == 7:
                yield node

    @classmethod
    def root(cls):
        return Atc("/", "ROOT", None)


class AtcTree:
    def __init__(self, root: Atc, lookup: dict):
        self.root = root
        self._lookup = lookup

    def get(self) -> Atc:
        return self._lookup[Atc]

    def nodes(self) -> Set[Atc]:
        return set(self._lookup.values())

    def dfs(self) -> Iterator[Atc]:
        yield from self.root.dfs()

    def bfs(self) -> Iterator[Atc]:
        yield from self.root.bfs()

    def leaves(self) -> Iterator[Atc]:
        yield from self.root.leaves()

    def __len__(self):
        return len(self.nodes())

    def __repr__(self):
        return f"AtcTree({len(self._lookup)} nodes @ {hex(id(self))})"

    def __str__(self):
        return repr(self)


class AtcParser:

    URL = "https://pubchem.ncbi.nlm.nih.gov/rest/pug_view/annotations/heading/JSON/?source=WHO%20ATC&heading=ATC+Code&response_type=save&response_basename=PubChemAnnotations_source=WHO%20ATC&heading=ATC+Code&page={page}"

    def __init__(self, cache_dir: PathLike = ".atc"):
        self.cache_dir = Path(cache_dir)
        self.cache_dir.mkdir(parents=True, exist_ok=True)

    def load(self, force_download: bool = False) -> AtcTree:
        atcs = {"/": Atc.root()}
        if not force_download and Path(self.cache_dir, "is-done").exists():
            logger.debug("Loading ATC tree from cache...")
            self._load_from_cache(atcs)
        else:
            logger.info("Downloading ATC tree from pubchem.ncbi.nlm.nih.gov ...")
            self._download(atcs)
        logger.debug(f"Loaded {len(atcs)} ATC codes")
        for atc in atcs.values():
            atc.children = sorted(list(atc.children))
        return AtcTree(atcs["/"], atcs)

    def _load_from_cache(self, atcs: dict):
        i = 1
        while True:
            p = self.cache_dir / f"page-{i}.txt"
            if p.exists():
                with open(p) as f:
                    data = json.loads(f.read())["Annotations"]["Annotation"]
                    for atc in self._parse(data, atcs):
                        pass
                i += 1

    def _download(self, atcs: dict):
        response = requests.get(AtcParser.URL.format(page=1))
        data = response.json()["Annotations"]
        n_pages = data["TotalPages"]
        for i in range(1, n_pages + 1):  # downloading page 1 twice, but whatever
            response = requests.get(AtcParser.URL.format(page=i))
            logger.debug(f"Downloading page {i} of {n_pages}.")
            (self.cache_dir / f"page-{i}.txt").write_text(response.text, encoding="utf-8")
            data = response.json()["Annotations"]
            for atc in self._parse(data["Annotation"], atcs):
                pass
            (self.cache_dir / ".is_done").write_text(str(datetime.now()), encoding="utf-8")

    def _parse(self, items: list, atcs: dict):
        pat = regex.compile(r"^(?:<[^>]+>)? *([^ ]+) +- +(?:<[^>]+>)? *([^<]+).*$", flags=regex.V1)
        root = atcs["/"]
        for item in items:
            for v0 in [
                _["Value"]["String"][0] for _ in item["Data"] if _["TOCHeading"] == "ATC Code"
            ]:
                parent = root
                for v in v0.split("<br>"):
                    m = pat.match(v.strip())
                    code, name = m.group(1).strip(), m.group(2).strip()
                    contains = code in atcs
                    if code in atcs:
                        child = atcs[code]
                    else:
                        child = Atc(code, name, parent)
                    parent.children.add(child)
                    parent = child
                    if not contains:
                        atcs[child.code] = child
                        yield child
                # and it resets parent each loop

    def __repr__(self):
        return self.__name__

    def __str__(self):
        return repr(self)


__all__ = ["Atc", "AtcTree", "AtcParser"]



from __future__ import annotations

import abc
import contextlib
import logging
import shutil
from pathlib import Path
from typing import Any, TypeVar, Union

import requests

# noinspection PyUnresolvedReferences
from pocketutils.core import PathLike, PathLikeUtils

T = TypeVar("T", covariant=True)
Y = TypeVar("Y")
Z = TypeVar("Z")
logger = logging.getLogger("pocketutils")


class Writeable(metaclass=abc.ABCMeta):
    @classmethod
    def isinstance(cls, value: Any):
        return hasattr(value, "write") and hasattr(value, "flush") and hasattr(value, "close")

    def write(self, msg):
        raise NotImplementedError()

    def flush(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class DevNull(Writeable):
    """Pretends to write but doesn't."""

    def write(self, msg):
        pass

    def flush(self):
        pass

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class LogWriter:
    """
    A call to a logger at some level, pretending to be a writer.
    Has a write method, as well as flush and close methods that do nothing.
    """

    def __init__(self, level: Union[int, str]):
        if isinstance(level, str):
            level = level.upper()
        self.level = logging.getLevelName(level)

    def write(self, msg: str):
        getattr(logger, self.level)(msg)

    def flush(self):
        pass

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class DelegatingWriter(object):
    # we CANNOT override TextIOBase: It causes hangs
    def __init__(self, *writers):
        self._writers = writers

    def write(self, s):
        for writer in self._writers:
            writer.write(s)

    def flush(self):
        for writer in self._writers:
            writer.flush()

    def close(self):
        for writer in self._writers:
            writer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class Capture:
    """
    A lazy string-like object that wraps around a StringIO result.
    It's too hard to fully subclass a string while keeping it lazy.
    """

    def __init__(self, cio):
        self.__cio = cio

    @property
    def lines(self):
        return self.split("\n")

    @property
    def value(self):
        return self.__cio.getvalue()

    def __repr__(self):
        return self.__cio.getvalue()

    def __str__(self):
        return self.__cio.getvalue()

    def __len__(self):
        return len(repr(self))

    def split(self, x: str):
        return self.__cio.getvalue().split(x)


class OpenMode(str):
    """
    Extended file open modes with a superset of meanings.
    The underlying string contains a Python open()-compatible string.

        - 'r' means read
        - 'w' and 'o' both mean overwrite
        - 'a' means append
        - 's' means "safe" -- complain if it exists (neither overwrite nor append)
        - 'b' means binary
        - 'z' means compressed with gzip; works in both binary and text modes
        - 'd' means detect gzip
    """

    # noinspection PyMissingConstructor
    def __init__(self, mode: str):
        self._raw = mode.replace("w", "o")
        self.internal = (
            self._raw.replace("o", "w")
            .replace("s", "w")
            .replace("z", "")
            .replace("i", "")
            .replace("d", "")
        )

    def __repr__(self):
        return self.internal

    def __str__(self):
        return self.internal

    @property
    def read(self) -> bool:
        return "r" in self._raw

    @property
    def write(self) -> bool:
        return "r" not in self._raw

    @property
    def safe(self) -> bool:
        return "s" in self._raw

    @property
    def overwrite(self) -> bool:
        return "o" in self._raw or "w" in self._raw

    @property
    def ignore(self) -> bool:
        return "i" in self._raw

    @property
    def append(self) -> bool:
        return "a" in self._raw

    @property
    def text(self) -> bool:
        return "b" not in self._raw

    @property
    def binary(self) -> bool:
        return "b" in self._raw

    @property
    def gzipped(self) -> bool:
        return "z" in self._raw

    def __eq__(self, other):
        if isinstance(other, OpenMode):
            return self._raw == other._raw
        elif isinstance(other, str):
            return self._raw == other.replace("w", "o")
        return False


def null_context():
    yield


@contextlib.contextmanager
def silenced(no_stdout: bool = True, no_stderr: bool = True):
    with contextlib.redirect_stdout(DevNull()) if no_stdout else null_context():
        with contextlib.redirect_stderr(DevNull()) if no_stderr else null_context():
            yield


def stream_download(url: str, path: PathLike):
    with requests.get(url, stream=True) as r:
        with Path(path).open("wb") as f:
            shutil.copyfileobj(r.raw, f)


__all__ = [
    "Writeable",
    "DevNull",
    "LogWriter",
    "DelegatingWriter",
    "Capture",
    "OpenMode",
    "PathLike",
]



from typing import Optional

import pandas as pd
import regex
from IPython.display import HTML, Markdown, display

_color_pattern = regex.compile("#?[A-Z0-9a-z]{6}", flags=regex.V1)


# noinspection PyTypeChecker
class J:
    """
    Convenience user tools to display HTML text in Jupyter notebooks.
    """

    from IPython.display import clear_output, display
    from IPython.utils import io as __io

    RED, GREEN, BLUE, PURPLE, CYAN, BOLD = (
        "#cc0000",
        "#008800",
        "#0000cc",
        "#880099",
        "#007777",
        "600",
    )

    @classmethod
    def full_width(cls, percent: bool = 100):
        display(HTML("<style>.container { width:" + str(percent) + "% !important; }</style>"))

    @classmethod
    def full(cls, df: pd.DataFrame) -> None:
        with pd.option_context("display.max_rows", 10 ** 10):
            with pd.option_context("display.max_columns", 10 ** 10):
                display(df)

    @classmethod
    def confirm(cls, msg: Optional[str] = None) -> bool:
        # return Tools.confirm(lambda: cls.blue("Confirm? [yes/no]" if msg is None else msg))
        cls.bold_colored("Confirm? [yes/no]" if msg is None else msg, cls.BLUE)
        return cls.__io.ask_yes_no("")

    @classmethod
    def hyperlink(cls, text: str, url: str) -> None:
        display(HTML(f'<a href="{url}">{text}</a>'))

    @classmethod
    def red(cls, text: str) -> None:
        cls.colored(text, cls.RED)

    @classmethod
    def blue(cls, text: str) -> None:
        cls.colored(text, cls.BLUE)

    @classmethod
    def green(cls, text: str) -> None:
        cls.colored(text, cls.GREEN)

    @classmethod
    def bold(cls, text: str) -> None:
        cls.styled(text, font_weight=cls.BOLD)

    @classmethod
    def colored(cls, text: str, color: str, **kwargs) -> None:
        cls.styled(text, color=cls._color(color), **kwargs)

    @classmethod
    def bold_colored(cls, text: str, color: str) -> None:
        cls.styled(text, color=color, font_weight=cls.BOLD)

    @classmethod
    def code(cls, text: str, **kwargs) -> None:
        cls.styled(text, white_space="pre", font_family="monospace", **kwargs)

    @classmethod
    def md(cls, text: str) -> None:
        display(Markdown(text))

    @classmethod
    def html(cls, text: str) -> None:
        display(HTML(text))

    @classmethod
    def styled(cls, text: str, *main, **css: str) -> None:
        """
        Display a span element styled with free CSS arguments.

        Args:
            text: The text to include in an HTML span
            main: Any free text of CSS, or a list of parameters
            css: Any free dict of CSS style parameters; spaces and underscores in the keys will be replaced with hypens
        """
        main = ";".join(main)
        css = ";".join(
            "{}: {}".format(k.replace("_", "-").replace(" ", "-"), str(v)) for k, v in css.items()
        )
        if len(main) > 0 and not main.endswith(";"):
            main += ";"
        z = (main + css) if (main + css).endswith(";") else (main + css + ";")
        display(HTML('<span style="{}">{}</span>'.format(z, text)))

    @classmethod
    def _color(cls, color: str) -> str:
        if _color_pattern.fullmatch(color) is None:
            raise ValueError(f"Invalid hex color {color}")
        return color if color.startswith("#") else "#" + color


class JFonts:
    """
    Renders HTML for matplotlib fonts.
    """

    @classmethod
    def one(cls, name: str) -> None:
        """
        Shows a single typeface as itself. Ex; will show Helvetica in Helvetica.
        """
        J.html(f'<p style="font-family:{name};">{name}</p>')

    @classmethod
    def mine(cls) -> None:
        """
        Shows all typefaces currently in the matplotlib rcParams.
        Each typeface is rendered as itself. Ex; will show Helvetica in Helvetica.
        Each font family (``plt.rcParams['font.family']``) is shown separately, with all its fonts underneath.
        """
        import matplotlib.pyplot as plt

        for family in plt.rcParams["font.family"]:
            items = "\n".join(
                [
                    f'<li style="font-family:{family};">{family}</li>'
                    for font in plt.rcParams.get("font." + family, [])
                ]
            )
            J.html(
                (
                    '<h4 style="padding-bottom:0;margin-bottom:0;">{}:</h4>\n'
                    '<ul style="padding-top:0;margin-top:0;margin-bottom:0;margin-top:0;">\n'
                    "{}\n</ul>"
                ).format(family, items)
            )

    @classmethod
    def every(cls, n_cols: int = 4) -> None:
        """
        Shows an HTML table of all typefaces rendered as themselves. Ex; will show Helvetica in Helvetica.
        Displays as an HTML table with ``n_cols``.
        Thanks to https://jonathansoma.com/lede/data-studio/matplotlib/list-all-fonts-available-in-matplotlib-plus-samples/.

        Args:
            n_cols: Number of columns in the created HTML table
        """
        from matplotlib.font_manager import fontManager

        def _show_font(name: str):
            return f"<p style='font-family:{name};'>{name}</p>"

        fonts = set([f.name for f in fontManager.ttflist])
        code = "\n".join([_show_font(font) for font in sorted(fonts)])
        J.html(f'<div style="column-count: {n_cols};">{code}</div>')


__all__ = ["J", "JFonts"]



from __future__ import annotations

import regex

from pocketutils.tools.string_tools import StringTools
from pocketutils.tools.unit_tools import UnitTools


class RefDims(dict):
    """
    Reference widths and heights by name.

    Example:
        Ex::

            widths[1"]
            widths['1/3 2_col']
            widths['(2/1) 2_col']

    .. note::
        If ``width_pad`` (``height_pad``, respectively) is set, this will be included in the calculation.
        For example, (1/3) 2_col will subtract off the appropriate padding for 3 cols (2 spaces shared between 3 columns).
        The size is then rounded to sigfigs after -- normally 6 sigfigs, but overridden with width_sigfigs / height_sigfigs.
        The sum / difference is applied after.

    """

    def __init__(self, axis: str, n_sigfigs: int = 6):
        super().__init__()
        self.axis = axis
        self.n_sigfigs = n_sigfigs
        self.current_text: str = ""
        self.current: float = 0.0

    def point(self, item: str) -> float:
        """
        Returns the resulting value (calling ``self[item]``), also setting the text and value pointers in-place.
        It's best for this to match plt.rcParams[figure.figsize], but this is not required.
        """
        scale = self[item]
        self.current_text = item
        self.current = scale
        return scale

    def __getitem__(self, item: str) -> float:
        # first, match exact inches
        # If name is empty or double quote, assume it means inch (ex: '1' is 1 inch; 1/3 is 1/3 inch)
        # And in this case, ignore padding (take the size as-is, except for rounding)
        # For ex, '1/3 "' gets converted to 0.333333 if self.n_sigfigs==6
        # Whereas if the user defined 'inch', then '1/3 inch' will get padding removed, so it might be, say, round(1/3 - 2*0.1/3)
        item = str(item)
        pat = regex.compile(r'^ *([0-9.]+) *"? *$', flags=regex.V1)
        match = pat.fullmatch(item.lower())
        if match is not None:
            return float(match.group(1))
        # now match fancy expressions
        # they're of the form (<numerator>/<denominator>?)? <name>
        # note the lazy ?? here
        pat = regex.compile(
            r"^ *(?:([0-9]+)(?:/([0-9]+))?)?? *([0-9A-Za-z_\-/]+) *$", flags=regex.V1
        )
        match = pat.fullmatch(item.lower())
        pad = self.get("pad", 0.0)
        numerator = int(match.group(1)) if match.group(1) is not None else 1
        denominator = int(match.group(2)) if match.group(2) is not None else 1
        name = match.group(3)
        name = StringTools.strip_off_start(StringTools.strip_off_start(name, "width_"), "height_")
        name_val = super().__getitem__(name)
        frac = denominator / numerator  # flip for simplicity
        # Note the math here:
        # Subtract off the padding needed
        # If it's it's 1/3 of some value, we'll need to add 2 pieces of padding shared between the three
        # This is then 2/3*padding per unit ==> so subtract it
        return UnitTools.round_to_sigfigs(name_val / frac - (frac - 1) * pad / frac, self.n_sigfigs)

    def __setitem__(self, key: str, value: float):
        try:
            super().__setitem__(key, float(value))
        except Exception:
            raise TypeError(f"{value} is not a float")


__all__ = ["RefDims"]



import logging
import math
from typing import Optional, SupportsFloat, Tuple, Union

import regex

from pocketutils.core._internal import nicesize
from pocketutils.core.exceptions import OutOfRangeError, StringPatternError
from pocketutils.tools.base_tools import BaseTools
from pocketutils.tools.string_tools import StringTools

logger = logging.getLogger("pocketutils")


class UnitTools(BaseTools):
    @classmethod
    def delta_time_to_str(cls, delta_sec: float, space: str = "") -> str:
        """
        Returns a pretty string from a difference in time in seconds.
        Rounds hours and minutes to 2 decimal places, and seconds to 1.
        Ex: delta_time_to_str(313) == 5.22min
            delta_sec: The time in seconds
            space: Space char between digits and units;
                good choices are empty, ASCII space, Chars.narrownbsp, Chars.thinspace, and Chars.nbsp.

        Returns:
            A string with units 'hr', 'min', or 's'
        """
        if abs(delta_sec) > 60 * 60 * 3:
            return (
                StringTools.strip_empty_decimal(str(round(delta_sec / 60 / 60, 2))) + space + "hr"
            )
        elif abs(delta_sec) > 180:
            return StringTools.strip_empty_decimal(str(round(delta_sec / 60, 2))) + space + "min"
        else:
            return StringTools.strip_empty_decimal(str(round(delta_sec, 1))) + space + "s"

    @classmethod
    def ms_to_minsec(cls, ms: int, space: str = "") -> str:
        """
        Converts a number of milliseconds to one of the following formats:
            - 10ms         if < 1 sec
            - 10:15        if < 1 hour
            - 10:15:33     if < 1 day
            - 5d:10:15:33  if > 1 day
        Prepends a minus sign (−) if negative.

        Args:
            ms: The milliseconds
            space: Space char between digits and 'ms' (if used);
                   good choices are empty, ASCII space, Chars.narrownbsp,
                   Chars.thinspace, and Chars.nbsp.

        Returns:
            A string of one of the formats above
        """
        ms = abs(int(ms))
        seconds = int((ms / 1000) % 60)
        minutes = int((ms / (1000 * 60)) % 60)
        hours = int((ms / (1000 * 60 * 60)) % 24)
        days = int(ms / (1000 * 60 * 60 * 24))
        z_hr = str(hours).zfill(2)
        z_min = str(minutes).zfill(2)
        z_sec = str(seconds).zfill(2)
        sgn = "−" if ms < 0 else ""
        if ms < 1000:
            return f"{sgn}{ms}{space}ms"
        elif days > 1:
            return f"{days}d:{z_hr}:{z_min}:{z_sec}"
        elif hours > 1:
            return f"{sgn}{z_hr}:{z_min}:{z_sec}"
        else:
            return f"{sgn}{z_min}:{z_sec}"

    @classmethod
    def friendly_size(cls, n_bytes: int, *, space: str = " ") -> str:
        """
        Returns a text representation of a number of bytes.
        Uses base 2 with IEC 1998, rounded to 0 decimal places, and without a space.
        """
        return nicesize(n_bytes, space=space)

    @classmethod
    def round_to_sigfigs(cls, num: SupportsFloat, sig_figs: int) -> float:
        """
        Round to specified number of sigfigs.

        Args:
            num: A Python or Numpy float or something that supports __float__
            sig_figs: The number of significant figures, non-negative

        Returns:
            A Python integer
        """
        if sig_figs < 0:
            raise OutOfRangeError(f"sig_figs {sig_figs} is negative", minimum=0)
        num = float(num)
        if num != 0:
            digits = -int(math.floor(math.log10(abs(num))) - (sig_figs - 1))
            return round(num, digits)
        else:
            return 0  # can't take the log of 0

    @classmethod
    def format_micromolar(
        cls,
        micromolar: float,
        n_sigfigs: Optional[int] = 5,
        *,
        adjust_units: bool = True,
        use_sigfigs: bool = True,
        space: str = "",
    ) -> str:
        """
        Returns a concentration with units, with the units scaled as needed.
        Can handle millimolar, micromolar, nanomolar, and picomolar.

        Args:
            micromolar: Value
            n_sigfigs: For rounding; no rounding if None
            adjust_units: If False, will always use micromolar
            use_sigfigs: If True, rounds to a number of significant figures; otherwise round to decimal places
            space: Space char between digits and units;
                   good choices are empty, ASCII space, Chars.narrownbsp, Chars.thinspace, and Chars.nbsp.

        Returns:
            The concentration with a suffix of µM, mM, nM, or mM
        """
        d = micromolar
        m = abs(d)
        unit = "µM"
        if adjust_units:
            if m < 1e-6:
                d *= 1e9
                unit = "fM"
            elif m < 1e-3:
                d *= 1e6
                unit = "pM"
            elif m < 1:
                d *= 1e3
                unit = "nM"
            elif m >= 1e6:
                d /= 1e6
                unit = "M"
            elif m >= 1e3:
                d /= 1e3
                unit = "mM"
        if n_sigfigs is None:
            pass
        elif use_sigfigs:
            d = cls.round_to_sigfigs(d, n_sigfigs)
        else:
            d = round(d, n_sigfigs)
        if round(d) == d and str(d).endswith(".0"):
            return str(d)[:-2] + space + unit
        else:
            return str(d) + space + unit

    @classmethod
    def split_species_micromolar(cls, text: str) -> Tuple[str, Optional[float]]:
        """
        Splits a name into a chemical/concentration pair, falling back with the full name.
        Ex: "abc 3.5uM" → (abc, 3.5)
        Ex: "abc 3.5 µM" → (abc, 3.5)
        Ex: "abc (3.5mM)" → (abc, 3500.0)
        Ex: "abc 3.5mM" → (abc, None)
        Ex: "3.5mM" → (3.5mM, None)  # an edge case: don't pass in only units
        Uses a moderately strict pattern for the drug and dose:
            - The dose must terminate the string, except for end parenthesis or whitespace.
            - The drug and dose must be separated by at least one non-alphanumeric, non-dot, non-hyphen character.
            - Units must follow the digits, separated by at most whitespace, and are case-sensitive.
        """
        # note the lazy ops in the first group and in the non-(alphanumeric/dot/dash) separator between the drug and dose
        pat = regex.compile(
            r"^\s*(.*?)(?:[^A-Za-z0-9.\-]+?[\s(\[{]*(\d+(?:.\d*)?)\s*([mµunpf]M)\s*[)\]}]*)?\s*$",
            flags=regex.V1,
        )
        match = pat.fullmatch(text)
        if match is None:
            raise StringPatternError(f"Text {text} couldn't be parsed", value=text, pattern=pat)
        if match.group(2) is None:
            return text.strip(), None
        else:
            drug = match.group(1).strip("([{)]}")
            dose = UnitTools.concentration_to_micromolar(float(match.group(2)), match.group(3))
            return drug, dose

    @classmethod
    def extract_micromolar(cls, text: str) -> Optional[float]:
        """
        Returns what looks like a concentration with units. Accepts one of: mM, µM, uM, nM, pM.
        Searches pretty flexibly.
        If no matches are found, returns None.
        If multiple matches are found, warns and returns None.
        """
        # we need to make sure mM ex isn't part of a larger name
        pat1 = regex.compile(r"(\d+(?:.\d*)?)\s*([mµunpf]M)\s*[)\]}]*", flags=regex.V1)

        def find(pat):
            return {
                UnitTools.concentration_to_micromolar(float(match.group(1)), match.group(2))
                for match in pat.finditer(text)
                if match is not None
            }

        matches = find(pat1)
        if len(matches) == 1:
            return next(iter(matches))
        elif len(matches) > 1:
            logger.warning(f"Found {len(matches)} potential doses: {matches} . Returning None.")
        return None

    @classmethod
    def concentration_to_micromolar(cls, digits: Union[str, float], units: str) -> float:
        """
        Ex: concentration_to_micromolar(53, 'nM')  # returns 0.053
        """
        return float(digits) * {
            "M": 1e6,
            "mM": 1e3,
            "µM": 1,
            "uM": 1,
            "nM": 1e-3,
            "pM": 1e-6,
            "fM": 1e-9,
        }[units]


__all__ = ["UnitTools"]



from __future__ import annotations

import warnings
from pathlib import Path
from typing import AbstractSet

from pocketutils.core.exceptions import NotConstructableError


class WarningsConfig:
    """
    Convenient API to add warning filters.
    Also provides :meth:`simplify_format`, which sets a less-verbose warning formatter.

    Example:
        >>> (
                GlobalWarningUtils.simplify_format()
                .filter_common()
                .never("Number of features differ")
            )
    """

    def __init__(self):
        raise NotConstructableError(f"Do not instantiate {self.__class__.__name__}")

    @classmethod
    def simplify_format(cls) -> __qualname__:
        """
        Common initialization, including setting a better formatter that doesn't say "WARNING:py.warnings:".
        """

        def new_formatter(message, category, filename, lineno, line=None):
            cat = category.__name__.replace("Warning", "")
            s = f"{Path(filename).name}:{lineno}: {cat}: {message}"
            return s.replace("WARNING:py.warnings:", "")

        warnings.formatwarning = new_formatter
        return cls

    @classmethod
    def filter(cls, **kwargs) -> __qualname__:
        """Same as warnings.filterwarnings."""
        warnings.filterwarnings(**kwargs)
        return cls

    @classmethod
    def never(cls, *substrings: str) -> __qualname__:
        """Adds a filter containing this substring, never showing the warning."""
        for substring in substrings:
            warnings.filterwarnings(message=".*?" + substring + ".*", action="ignore")
        return cls

    @classmethod
    def once(cls, *substrings: str) -> __qualname__:
        """Adds a filter containing this substring, warning only once."""
        for substring in substrings:
            warnings.filterwarnings(message=".*?" + substring + ".*", action="once")
        return cls

    @classmethod
    def filter_common(cls) -> __qualname__:
        """
        Adds filters for common unavoidable warnings from numpy, pandas, scikit-learn, etc.

        See ``common_never_substrings`` and ``common_once_substrings``.
        """
        return cls.never(*cls.common_never()).once(*cls.common_once())

    @classmethod
    def common_never(cls) -> AbstractSet[str]:
        return {
            "libuv only supports millisecond timer resolution",
            "or '1type' as a synonym of type is deprecated",
            "Series.nonzero() is deprecated and will be removed in a future version",
            "Monkey-patching ssl after ssl has already been imported may lead to errors",
            "your performance may suffer as PyTables will pickle object types that it cannot map directly to c-types",
        }

    @classmethod
    def common_once(cls) -> AbstractSet[str]:
        return {
            "Trying to unpickle estimator",
        }


__all__ = ["WarningsConfig"]



import json
import warnings
from copy import copy
from typing import Any, Callable, Iterable, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import regex

from pocketutils.core import JsonEncoder
from pocketutils.core.chars import *
from pocketutils.core.exceptions import OutOfRangeError
from pocketutils.tools.base_tools import BaseTools

T = TypeVar("T")
V = TypeVar("V")
_control_chars = regex.compile(r"\p{C}", flags=regex.V1)


class StringTools(BaseTools):
    @classmethod
    def pretty_dict(cls, dct: Mapping[Any, Any]) -> str:
        """
        Returns a pretty-printed dict, complete with indentation. Will fail on non-JSON-serializable datatypes.
        """
        # return Pretty.condensed(dct)
        return json.dumps(
            dct,
            default=JsonEncoder().default,
            sort_keys=True,
            indent=2,
            ensure_ascii=False,
        )

    @classmethod
    def extract_group_1(
        cls, pattern: Union[str, regex.Pattern], value: Optional[str], ignore_null: bool = False
    ) -> Optional[str]:
        warnings.warn("extract_group_1 will be removed", DeprecationWarning)
        """
        Performs a ``fullmatch`` on a target string and returns capture group 1, or None if there was no match.

        Args:
            pattern: Regex pattern
            value: The target string
            ignore_null: If True, returns None if ``value`` is None; otherwise raises a ValueError if ``value`` is None
                         (Useful for *map*-like operations.)

        Returns The first capture group, or None
        """
        pattern = (
            pattern
            if isinstance(pattern, regex.Pattern)
            else regex.compile(pattern, flags=regex.V1)
        )
        if pattern.groups != 1:
            raise ValueError(f"Pattern {pattern} has {pattern.groups} groups, not 1")
        if value is None and ignore_null:
            return None
        match = pattern.fullmatch(value)
        if match is None:
            return None
        return match.group(1)

    @classmethod
    def join_to_str(cls, *items: Any, last: str, sep: str = ", ") -> str:
        """
        Joins items to something like "cat, dog, and pigeon" or "cat, dog, or pigeon".

        Args:
            *items: Items to join; ``str(item) for item in items`` will be used
            last: Probably "and", "or", "and/or", or ""
                    Spaces are added/removed as needed if ``suffix`` is alphanumeric
                    or "and/or", after stripping whitespace off the ends.
            sep: Used to separate all words; include spaces as desired

        Examples:
            - ``join_to_str(["cat", "dog", "elephant"], last="and")  # cat, dog, and elephant``
            - ``join_to_str(["cat", "dog"], last="and")  # cat and dog``
            - ``join_to_str(["cat", "dog", "elephant"], last="", sep="/")  # cat/dog/elephant``
        """
        if last.strip().isalpha() or last.strip() == "and/or":
            last = last.strip() + " "
        items = [str(s).strip("'" + '"' + " ") for s in items]
        if len(items) > 2:
            return sep.join(items[:-1]) + sep + last + items[-1]
        else:
            return (" " + last + " ").join(items)

    @classmethod
    def strip_control_chars(cls, s: str) -> str:
        """
        Strips all characters under the Unicode 'Cc' category.
        """
        return _control_chars.sub("", s)

    @classmethod
    def roman_to_arabic(
        cls, roman: str, min_val: Optional[int] = None, max_val: Optional[int] = None
    ) -> int:
        """
        Converts roman numerals to an integer.

        Args:
            roman: A string like "MCIV"
            min_val: Raise a ValueError if the parsed value is less than this
            max_val: Raise a ValueError if the parsed value is more than this

        Returns:
            The arabic numeral as a Python int
        """
        # this order is IMPORTANT!
        mp = dict(
            IV=4, IX=9, XL=40, XC=90, CD=400, CM=900, I=1, V=5, X=10, L=50, C=100, D=500, M=1000
        )
        for k, v in mp.items():
            roman = roman.replace(k, str(v))
        # it'll just error if it's empty
        try:
            value = sum((int(num) for num in roman))
        except (ValueError, StopIteration):
            raise ValueError(f"Cannot parse roman numerals '{roman}'")
        if min_val is not None and value < min_val or min_val is not None and roman > max_val:
            raise ValueError(f"Value {roman} (int={value}) is out of range ({min_val}, {max_val})")
        return value

    @classmethod
    def retab(cls, s: str, nspaces: int) -> str:
        """
        Converts indentation with spaces to tab indentation.

        Args:
            s: The string to convert
            nspaces: A tab is this number of spaces
        """

        def fix(m):
            n = len(m.group(1)) // nspaces
            return "\t" * n + " " * (len(m.group(1)) % nspaces)

        return regex.sub("^( +)", fix, s, flags=regex.V1 | regex.MULTILINE)

    @classmethod
    def strip_empty_decimal(cls, num: Union[float, str]) -> str:
        """
        Replaces prefix . with 0. and strips trailing .0 and trailing .
        """
        try:
            float(num)
        except TypeError:
            if not isinstance(num, str):
                raise TypeError("Must be either str or float-like") from None
        t = str(num)
        if t.startswith("."):
            t = "0" + t
        if "." in t:
            return t.rstrip("0").rstrip(".")
        else:
            return t

    @classmethod
    def tabs_to_list(cls, s: str) -> Sequence[str]:
        """
        Splits by tabs, but preserving quoted tabs, stripping quotes.
        In other words, will not split within a quoted substring.
        Double and single quotes are handled.
        """
        pat = regex.compile(r"""((?:[^\t"']|"[^"]*"|'[^']*')+)""", flags=regex.V1)
        # Don't strip double 2x quotes: ex ""55"" should be "55", not 55
        def strip(i: str) -> str:
            if i.endswith('"') or i.endswith("'"):
                i = i[:-1]
            if i.startswith('"') or i.startswith("'"):
                i = i[1:]
            return i.strip()

        return [strip(i) for i in pat.findall(s)]

    @classmethod
    def truncate(
        cls,
        s: Optional[str],
        n: int,
        *,
        null: Optional[str] = None,
    ) -> Optional[str]:
        """
        Returns a string if it has ``n`` or fewer characters;
        otherwise truncates to length ``n-1`` and appends ``…`` (UTF character).
        If ``s`` is None and ``always_dots`` is True, returns ``n`` copies of ``.`` (as a string).
        If ``s`` is None otherwise, returns None.

        Args:
            s: The string
            n: The maximum length, inclusive
            null: Replace ``None`` with this string

        Returns:
            A string or None
        """
        if s is None:
            return null
        if len(s) > n:
            nx = max(0, n - 1)
            return s[:nx] + "…"
        return s

    # these are provided to avoid having to call with labdas or functools.partial
    @classmethod
    def truncating(
        cls,
        n: int = 40,
        always_dots: bool = False,
        *,
        null: Optional[str] = None,
    ) -> Callable[[str], str]:
        # pretty much functools.partial
        def trunc(s: str) -> str:
            return cls.truncate(s, n, null=null)

        trunc.__name__ = f"truncate({n},{'…' if always_dots else ''})"
        return trunc

    @classmethod
    def longest(cls, parts: Iterable[T]) -> T:
        """
        Returns an element with the highest ``len``.
        """
        mx = ""
        for i, x in enumerate(parts):
            if len(x) > len(mx):
                mx = x
        return mx

    @classmethod
    def longest_str(cls, parts: Iterable[str]) -> str:
        """
        Returns the argmax by length.
        """
        return cls.longest(parts)

    @classmethod
    def strip_off_start(cls, s: str, pre: str) -> str:
        """
        Strips the full string ``pre`` from the start of ``str``.
        See ``Tools.strip_off`` for more info.
        """
        if not isinstance(pre, str):
            raise TypeError(f"{pre} is not a string")
        if s.startswith(pre):
            s = s[len(pre) :]
        return s

    @classmethod
    def strip_off_end(cls, s: str, suf: str) -> str:
        """
        Strips the full string ``suf`` from the end of ``str``.
        See `Tools.strip_off` for more info.
        """
        if not isinstance(suf, str):
            raise TypeError(f"{suf} is not a string")
        if s.endswith(suf):
            s = s[: -len(suf)]
        return s

    @classmethod
    def strip_off(cls, s: str, prefix_or_suffix: str) -> str:
        """
        Strip a substring from the beginning or end of a string (at most 1 occurrence).
        """
        return StringTools.strip_off_start(
            StringTools.strip_off_end(s, prefix_or_suffix), prefix_or_suffix
        )

    @classmethod
    def strip_ends(cls, s: str, prefix: str, suffix: str) -> str:
        """
        Strips a substring from the start, and another substring from the end, of a string (at most 1 occurrence each).
        """
        return StringTools.strip_off_start(StringTools.strip_off_end(s, suffix), prefix)

    @classmethod
    def strip_any_ends(
        cls,
        s: str,
        prefixes: Union[str, Sequence[str]],
        suffixes: Union[str, Sequence[str]],
    ) -> str:
        """
        Flexible variant that strips any number of prefixes and any number of suffixes.
        Also less type-safe than more specific variants.
        Note that the order of the prefixes (or suffixes) DOES matter.
        """
        prefixes = (
            [str(z) for z in prefixes]
            if StringTools.is_true_iterable(prefixes)
            else [str(prefixes)]
        )
        suffixes = (
            [str(z) for z in suffixes]
            if StringTools.is_true_iterable(suffixes)
            else [str(suffixes)]
        )
        s = str(s)
        for pre in prefixes:
            if s.startswith(pre):
                s = s[len(pre) :]
        for suf in suffixes:
            if s.endswith(suf):
                s = s[: -len(suf)]
        return s

    @classmethod
    def strip_brackets(cls, text: str) -> str:
        """
        Strips any and all pairs of brackets from start and end of a string, but only if they're paired.
        See Also:
             strip_paired
        """
        pieces = [
            ("(", ")"),
            ("[", "]"),
            ("[", "]"),
            ("{", "}"),
            ("<", ">"),
            (Chars.lshell, Chars.rshell),
            (Chars.langle, Chars.rangle),
            (Chars.ldparen, Chars.rdparen),
            (Chars.ldbracket, Chars.rdbracket),
            (Chars.ldangle, Chars.rdangle),
            (Chars.ldshell, Chars.rdshell),
        ]
        return StringTools.strip_paired(text, pieces)

    @classmethod
    def strip_quotes(cls, text: str) -> str:
        """
        Strips any and all pairs of quotes from start and end of a string, but only if they're paired.

        See Also:
            strip_paired
        """
        pieces = [
            ("`", "`"),
            (Chars.lsq, Chars.rsq),
            (Chars.ldq, Chars.rdq),
            ("'", "'"),
            ('"', '"'),
        ]
        return StringTools.strip_paired(text, pieces)

    @classmethod
    def strip_brackets_and_quotes(cls, text: str) -> str:
        """
        Strips any and all pairs of brackets and quotes from start and end of a string, but only if they're paired.

        See Also:
            strip_paired
        """
        pieces = [
            ("(", ")"),
            ("[", "]"),
            ("[", "]"),
            ("{", "}"),
            ("<", ">"),
            (Chars.lshell, Chars.rshell),
            (Chars.langle, Chars.rangle),
            ("`", "`"),
            (Chars.lsq, Chars.rsq),
            (Chars.ldq, Chars.rdq),
            ("'", "'"),
            ('"', '"'),
            (Chars.ldparen, Chars.rdparen),
            (Chars.ldbracket, Chars.rdbracket),
            (Chars.ldangle, Chars.rdangle),
            (Chars.ldshell, Chars.rdshell),
        ]
        return StringTools.strip_paired(text, pieces)

    @classmethod
    def strip_paired(cls, text: str, pieces: Iterable[Tuple[str, str]]) -> str:
        """
        Strips pairs of (start, end) from the ends of strings.

        Example:
            >>> StringTools.strip_paired("[(abc]", [("(", ")"), ("[", "]"))  # returns "(abc"

        Also see ``strip_brackets``
        """
        if any([a for a in pieces if len(a) != 2]):
            raise ValueError(f"Each item must be a string of length 2: (stard, end); got {pieces}")
        text = str(text)
        while len(text) > 0:
            yes = False
            for a, b in pieces:
                while text.startswith(a) and text.endswith(b):
                    text = text[1:-1]
                    yes = True
            if not yes:
                break
        return text

    @classmethod
    def replace_all(cls, s: str, rep: Mapping[str, str]) -> str:
        """
        Simply replace multiple things in a string.
        """
        warnings.warn("replace_all will be removed", DeprecationWarning)
        for k, v in rep.items():
            s = s.replace(k, v)
        return s

    @classmethod
    def superscript(cls, s: Union[str, float]) -> str:
        """
        Replaces digits, +, =, (, and ) with equivalent Unicode superscript chars (ex ¹).
        """
        return "".join(dict(zip("0123456789-+=()", "⁰¹²³⁴⁵⁶⁷⁸⁹⁻⁺⁼⁽⁾")).get(c, c) for c in s)

    @classmethod
    def subscript(cls, s: Union[str, float]) -> str:
        """
        Replaces digits, +, =, (, and ) with equivalent Unicode subscript chars (ex ₁).
        """
        return "".join(dict(zip("0123456789+-=()", "₀₁₂₃₄₅₆₇₈₉₊₋₌₍₎")).get(c, c) for c in s)

    @classmethod
    def unsuperscript(cls, s: Union[str, float]) -> str:
        """
        Replaces Unicode superscript digits, +, =, (, and ) with normal chars.
        """
        return "".join(dict(zip("⁰¹²³⁴⁵⁶⁷⁸⁹⁻⁺⁼⁽⁾", "0123456789-+=()")).get(c, c) for c in s)

    @classmethod
    def unsubscript(cls, s: Union[str, float]) -> str:
        """
        Replaces Unicode superscript digits, +, =, (, and ) with normal chars.
        """
        return "".join(dict(zip("₀₁₂₃₄₅₆₇₈₉₊₋₌₍₎", "0123456789+-=()")).get(c, c) for c in s)

    @classmethod
    def dashes_to_hm(cls, s: str) -> str:
        """
        Replaces most Latin-alphabet dash-like and hyphen-like characters with a hyphen-minus.
        """
        smallem = "﹘"
        smallhm = "﹣"
        fullhm = "－"
        for c in [
            Chars.em,
            Chars.en,
            Chars.fig,
            Chars.minus,
            Chars.hyphen,
            Chars.nbhyphen,
            smallem,
            smallhm,
            fullhm,
        ]:
            s = str(s).replace(c, "-")
        return s

    @classmethod
    def pretty_float(cls, v: Union[float, int], n_sigfigs: Optional[int] = 5) -> str:
        """
        Represents a float as a string, with symbols for NaN and infinity.
        The returned string always has a minus or + prepended. Strip off the plus with .lstrip('+').
        If v is an integer (by isinstance), makes sure to display without a decimal point.
        If n_sigfigs < 2, will never have a
        For ex:
            - StringTools.pretty_float(.2222222)       # '+0.22222'
            - StringTools.pretty_float(-.2222222)      # '−0.22222' (Unicode minus)
            - StringTools.pretty_float(-float('inf'))  # '−∞'
            - StringTools.pretty_float(np.NaN)         # '⌀'
        """
        # TODO this seems absurdly long for what it does
        if n_sigfigs is None or n_sigfigs < 1:
            raise OutOfRangeError(
                f"Sigfigs of {n_sigfigs} is nonpositive",
                value=n_sigfigs,
                minimum=1,
            )
        # first, handle NaN and infinities
        if np.isneginf(v):
            return Chars.minus + Chars.inf
        elif np.isposinf(v):
            return "+" + Chars.inf
        elif np.isnan(v):
            return Chars.null
        # sweet. it's a regular float or int.
        if n_sigfigs is None:
            s = StringTools.strip_empty_decimal(str(v))
        else:
            # yes, this is weird. we need to convert from str to float then back to str
            s = str(float(str(("%." + str(n_sigfigs) + "g") % v)))
        # remove the .0 if the precision doesn't support it
        # if v >= 1 and n_sigfigs<2, it couldn't have a decimal
        # and if n_sigfigs<1, it definitely can't
        # and ... %g does this.
        if isinstance(v, int) or n_sigfigs is not None and n_sigfigs < 2:
            s = StringTools.strip_empty_decimal(s)
        # prepend + or - (unless 0)
        if float(s) == 0.0:
            return s
        s = s.replace("-", Chars.minus)
        if not s.startswith(Chars.minus):
            s = "+" + s[1:]
        if len(s) > 1 and s[1] == ".":
            s = s[0] + "0." + s[2:]
        return s

    @classmethod
    def pretty_function(
        cls, function, *, with_address: bool = False, prefix: str = "⟨", suffix: str = "⟩"
    ) -> str:
        """
        Get a better and shorter name for a function than str(function).
        Ex: pprint_function(lambda s: s)  == '<λ>'
        - Instead of '<bound method ...', you'll get '<name(nargs)>'
        - Instead of 'lambda ...', you'll get '<λ(nargs)>'
        - etc.
        NOTE 1: If function is None, returns '⌀'
        NOTE 2: If function does not have __name__, returns prefix + type(function) + <address> + suffix
        NOTE 3: If it's a primitive, returns str(function)

        Args:
            function: Can be anything, but especially useful for functions
            with_address: Include ``@ hex-mem-addr`` in the name
            prefix: Prefix to the whole string
            suffix: Suffix to the whole string
        """
        if function is None:
            return Chars.null
        n_args = str(function.__code__.co_argcount) if hasattr(function, "__code__") else "?"
        pat = regex.compile(r"^<bound method [^ .]+\.([^ ]+) of (.+)>$", flags=regex.V1)
        boundmatch = pat.fullmatch(str(function))
        pat = regex.compile(r"<([A-Za-z0-9_.<>]+)[ ']*object", flags=regex.V1)
        objmatch = pat.search(str(function))  # instance of global or local class
        addr = " @ " + hex(id(function)) if with_address else ""
        if cls.is_lambda(function):
            # simplify lambda functions!
            return prefix + "λ(" + n_args + ")" + addr + suffix
        elif boundmatch is not None:
            # it's a method (bound function)
            # don't show the address of the instance AND its method
            pat = regex.compile(r"@ ?0x[0-9a-hA-H]+\)?$", flags=regex.V1)
            s = pat.sub("", boundmatch.group(2)).strip()
            return (
                prefix + "`" + s + "`." + boundmatch.group(1) + "(" + n_args + ")" + addr + suffix
            )
        elif isinstance(function, type):
            # it's a class
            return prefix + "type:" + function.__name__ + suffix
        elif callable(function):
            # it's an actual function
            return prefix + function.__name__ + addr + suffix
        elif hasattr(function, "__dict__") and len(function.__dict__) > 0:
            # it's a member with attributes
            # it's interesting enough that it may have a good __str__
            s = StringTools.strip_off_end(
                StringTools.strip_off_start(str(function), prefix), suffix
            )
            return prefix + s + addr + suffix
        elif objmatch is not None:
            # it's an instance without attributes
            s = objmatch.group(1)
            if "." in s:
                s = s[s.rindex(".") + 1 :]
            return prefix + s + addr + suffix
        else:
            # it's a primitive, etc
            s = StringTools.strip_off_end(
                StringTools.strip_off_start(str(function), prefix), suffix
            )
            return s

    @classmethod
    def greek_to_name(cls) -> Mapping[str, str]:
        """
        Returns a dict from Greek lowercase+uppercase Unicode chars to their full names
        @return A defensive copy
        """
        return copy(StringTools._greek_alphabet)

    @classmethod
    def name_to_greek(cls) -> Mapping[str, str]:
        """
        Returns a dict from Greek lowercase+uppercase letter names to their Unicode chars
        @return A defensive copy
        """
        return {v: k for k, v in StringTools._greek_alphabet.items()}

    @classmethod
    def fix_greek(cls, s: str, lowercase: bool = False) -> str:
        """
        Replaces Greek letter names with their Unicode equivalents.
        Does this correctly by replacing superstrings before substrings.
        Ex: '1-beta' is '1-β' rather than '1-bη'
        If lowercase is True: Replaces Beta, BeTa, and BETA with β
        Else: Replaces Beta with a capital Greek Beta and ignores BETA and BeTa.
        """
        # Clever if I may say so:
        # If we just sort from longest to shortest, we can't replace substrings by accident
        # For example we'll replace 'beta' before 'eta', so '1-beta' won't become '1-bη'
        greek = sorted(
            [(v, k) for k, v in StringTools._greek_alphabet.items()],
            key=lambda t: -len(t[1]),
        )
        for k, v in greek:
            if k[0].isupper() and lowercase:
                continue
            if lowercase:
                s = regex.compile(k, flags=regex.V1 | regex.IGNORECASE).sub(v, s)
            else:
                s = s.replace(k, v)
        return s

    @classmethod
    def join(
        cls,
        seq: Iterable[T],
        *,
        sep: str = "\t",
        attr: Optional[str] = None,
        prefix: str = "",
        suffix: str = "",
    ) -> str:
        """
        Join elements into a str more easily than ''.join. Just simplifies potentially long expressions.
        Won't break with ValueError if the elements aren't strs.
        Ex:
            - StringTools.join([1,2,3])  # "1    2    3"
            - StringTools.join(cars, sep=',', attr='make', prefix="(", suffix=")")`  # "(Ford),(Ford),(BMW)"

            seq: Sequence of elements
            sep: Delimiter
            attr: Get this attribute from each element (in `seq`), or use the element itself if None
            prefix: Prefix before each item
            suffix: Suffix after each item

        Returns:
            A string
        """
        if attr is None:
            return sep.join([prefix + str(s) + suffix for s in seq])
        else:
            return sep.join([prefix + str(getattr(s, attr)) + suffix for s in seq])

    @classmethod
    def join_kv(
        cls,
        seq: Mapping[T, V],
        *,
        sep: str = "\t",
        eq: str = "=",
        prefix: str = "",
        suffix: str = "",
    ) -> str:
        """
        Joins dict elements into a str like 'a=1, b=2, c=3`.
        Won't break with ValueError if the keys or values aren't strs.

        Args:
            seq: Dict-like, with ``items()``
            sep: Delimiter
            eq: Separates a key with its value
            prefix: Prepend before every key
            suffix: Append after every value

        Returns:
            A string
        """
        return sep.join([prefix + str(k) + eq + str(v) + suffix for k, v in seq.items()])

    _greek_alphabet = {
        "\u0391": "Alpha",
        "\u0392": "Beta",
        "\u0393": "Gamma",
        "\u0394": "Delta",
        "\u0395": "Epsilon",
        "\u0396": "Zeta",
        "\u0397": "Eta",
        "\u0398": "Theta",
        "\u0399": "Iota",
        "\u039A": "Kappa",
        "\u039B": "Lambda",
        "\u039C": "Mu",
        "\u039D": "Nu",
        "\u039E": "Xi",
        "\u039F": "Omicron",
        "\u03A0": "Pi",
        "\u03A1": "Rho",
        "\u03A3": "Sigma",
        "\u03A4": "Tau",
        "\u03A5": "Upsilon",
        "\u03A6": "Phi",
        "\u03A7": "Chi",
        "\u03A8": "Psi",
        "\u03A9": "Omega",
        "\u03B1": "alpha",
        "\u03B2": "beta",
        "\u03B3": "gamma",
        "\u03B4": "delta",
        "\u03B5": "epsilon",
        "\u03B6": "zeta",
        "\u03B7": "eta",
        "\u03B8": "theta",
        "\u03B9": "iota",
        "\u03BA": "kappa",
        "\u03BB": "lambda",
        "\u03BC": "mu",
        "\u03BD": "nu",
        "\u03BE": "xi",
        "\u03BF": "omicron",
        "\u03C0": "pi",
        "\u03C1": "rho",
        "\u03C3": "sigma",
        "\u03C4": "tau",
        "\u03C5": "upsilon",
        "\u03C6": "phi",
        "\u03C7": "chi",
        "\u03C8": "psi",
        "\u03C9": "omega",
    }


__all__ = ["StringTools"]



from dataclasses import dataclass
from typing import Union

from matplotlib.axes import Axes
from matplotlib.axis import Axis


@dataclass(frozen=True, repr=True, order=True)
class Coords:
    x0: float
    x1: float
    y0: float
    y1: float


@dataclass(frozen=True, repr=True, order=True)
class TickBounder:
    """
    Forces Axes limits to start and/or end at major or minor ticks.

    Each argument in the constructor can be:
    - ``True`` -- set the bound according to the ticks
    - ``False`` -- do not change the bound
    - a ``float`` - set to this

    Args:
        x0: left bound (``.getxlim()[0]``)
        y0: bottom bound (``.getylim()[0]``)
        x1: right bound (``.getxlim()[1]``)
        y1: top bound (``.getylim()[1]``)
        major: Use major tick marks rather than minor

    Example:
        This example will bound maximum width and height of the Axes
        to the smallest tick that fits the data,
        and will set the minimum width and height to 0::

            ticker = TickBounder(x=0, y0=0, x1=True, y1=True)
            ticker.adjust(ax)
    """

    x0: Union[bool, float] = True
    y0: Union[bool, float] = True
    x1: Union[bool, float] = True
    y1: Union[bool, float] = True
    major: bool = True

    def adjust(self, ax: Axes) -> Axes:
        coords = self.adjusted(ax)
        ax.set_xlim(coords.x0, coords.x1)
        ax.set_ylim(coords.y0, coords.y1)
        return ax

    def adjusted(self, ax: Axes) -> Coords:
        return Coords(
            x0=self._adjust(ax.xaxis, self.x0, ax.get_xlim()[0], 0),
            x1=self._adjust(ax.xaxis, self.x0, ax.get_xlim()[1], 1),
            y0=self._adjust(ax.yaxis, self.y0, ax.get_ylim()[0], 0),
            y1=self._adjust(ax.yaxis, self.y0, ax.get_ylim()[1], 1),
        )

    def _adjust(self, ax: Axis, to: Union[bool, float], default: float, side: int) -> float:
        if to is False:
            return default
        try:
            return float(to)
        except ValueError:
            pass
        vals = ax.get_majorticklocs() if self.major else ax.get_minorticklocs()
        return vals[side]


__all__ = ["TickBounder", "Coords"]



import gzip
import hashlib
import importlib.metadata
import json
import locale
import logging
import os
import platform
import shutil
import socket
import stat
import struct
import sys
import tempfile
import warnings
from contextlib import contextmanager
from datetime import datetime, timezone
from getpass import getuser
from pathlib import Path, PurePath
from typing import Any, Generator, Iterable, Mapping, Optional, Sequence, SupportsBytes, Type, Union

import numpy as np
import pandas as pd
import regex

from pocketutils.core import JsonEncoder
from pocketutils.core.exceptions import (
    AlreadyUsedError,
    ContradictoryRequestError,
    FileDoesNotExistError,
    ParsingError,
)
from pocketutils.core.hashers import *
from pocketutils.core.input_output import OpenMode, PathLike, Writeable
from pocketutils.core.web_resource import *
from pocketutils.tools.base_tools import BaseTools
from pocketutils.tools.path_tools import PathTools

logger = logging.getLogger("pocketutils")
COMPRESS_LEVEL = 9
ENCODING = "utf8"

try:
    import jsonpickle
    import jsonpickle.ext.numpy as jsonpickle_numpy

    jsonpickle_numpy.register_handlers()
    import jsonpickle.ext.pandas as jsonpickle_pandas

    jsonpickle_pandas.register_handlers()
except ImportError:
    # zero them all out
    jsonpickle, jsonpickle_numpy, jsonpickle_pandas = None, None, None
    logger.debug("Could not import jsonpickle", exc_info=True)


try:
    from defusedxml import ElementTree
except ImportError:
    logger.warning("Could not import defusedxml; falling back to xml")
    from xml.etree import ElementTree


class FilesysTools(BaseTools):
    """
    Tools for file/directory creation, etc.

    .. caution::
        Some functions may be insecure.
    """

    @classmethod
    def new_hasher(cls, algorithm: str = "sha1") -> Hasher:
        return Hasher(algorithm)

    @classmethod
    def new_webresource(
        cls, url: str, archive_member: Optional[str], local_path: PathLike
    ) -> WebResource:
        return WebResource(url, archive_member, local_path)

    @classmethod
    def is_linux(cls) -> bool:
        return sys.platform == "linux"

    @classmethod
    def is_windows(cls) -> bool:
        return sys.platform == "win32"

    @classmethod
    def is_macos(cls) -> bool:
        return sys.platform == "darwin"

    @classmethod
    def get_env_info(cls, *, include_insecure: bool = False) -> Mapping[str, str]:
        """
        Get a dictionary of some system and environment information.
        Includes os_release, hostname, username, mem + disk, shell, etc.

        Args:
            include_insecure: Include data like hostname and username

        .. caution ::
            Even with ``include_insecure=False``, avoid exposing this data to untrusted
            sources. For example, this includes the specific OS release, which could
            be used in attack.
        """
        try:
            import psutil
        except ImportError:
            psutil = None
            logger.warning("psutil is not installed, so cannot get extended env info")

        now = datetime.now(timezone.utc).astimezone().isoformat()
        uname = platform.uname()
        language_code, encoding = locale.getlocale()
        # build up this dict:
        data = {}

        def _try(os_fn, k: str, *args):
            if any((a is None for a in args)):
                return None
            try:
                v = os_fn(*args)
                data[k] = v
                return v
            except (OSError, ImportError):
                return None

        data.update(
            dict(
                platform=platform.platform(),
                python=".".join(str(i) for i in sys.version_info),
                os=uname.system,
                os_release=uname.release,
                os_version=uname.version,
                machine=uname.machine,
                byte_order=sys.byteorder,
                processor=uname.processor,
                build=sys.version,
                python_bits=8 * struct.calcsize("P"),
                environment_info_capture_datetime=now,
                encoding=encoding,
                locale=locale,
                recursion_limit=sys.getrecursionlimit(),
                float_info=sys.float_info,
                int_info=sys.int_info,
                flags=sys.flags,
                hash_info=sys.hash_info,
                implementation=sys.implementation,
                switch_interval=sys.getswitchinterval(),
                filesystem_encoding=sys.getfilesystemencoding(),
            )
        )
        if "LANG" in os.environ:
            data["lang"] = os.environ["LANG"]
        if "SHELL" in os.environ:
            data["shell"] = os.environ["SHELL"]
        if "LC_ALL" in os.environ:
            data["lc_all"] = os.environ["LC_ALL"]
        if hasattr(sys, "winver"):
            data["win_ver"] = (sys.getwindowsversion(),)
        if hasattr(sys, "macver"):
            data["mac_ver"] = (sys.mac_ver(),)
        if hasattr(sys, "linux_distribution"):
            data["linux_distribution"] = (sys.linux_distribution(),)
        if include_insecure:
            _try(getuser, "username")
            _try(os.getlogin, "login")
            _try(socket.gethostname, "hostname")
            _try(os.getcwd, "cwd")
            pid = _try(os.getpid, "pid")
            ppid = _try(os.getppid, "parent_pid")
            if hasattr(os, "getpriority"):
                _try(os.getpriority, "priority", os.PRIO_PROCESS, pid)
                _try(os.getpriority, "parent_priority", os.PRIO_PROCESS, ppid)
        if psutil is not None:
            data.update(
                dict(
                    disk_used=psutil.disk_usage(".").used,
                    disk_free=psutil.disk_usage(".").free,
                    memory_used=psutil.virtual_memory().used,
                    memory_available=psutil.virtual_memory().available,
                )
            )
        return {k: str(v) for k, v in dict(data).items()}

    @classmethod
    def list_package_versions(cls) -> Mapping[str, str]:
        """
        Returns installed packages and their version numbers.
        Reliable; uses importlib (Python 3.8+).
        """
        # calling .metadata reads the metadata file
        # and .version is an alias to .metadata["version"]
        # so make sure to only read once
        # TODO: get installed extras?
        dct = {}
        for d in importlib.metadata.distributions():
            meta = d.metadata
            dct[meta["name"]] = meta["version"]
        return dct

    @classmethod
    def delete_surefire(cls, path: PathLike) -> Optional[Exception]:
        """
        Deletes files or directories cross-platform, but working around multiple issues in Windows.

        Returns:
            None, or an Exception for minor warnings

        Raises:
            IOError: If it can't delete
        """
        # we need this because of Windows
        path = Path(path)
        logger.debug(f"Permanently deleting {path} ...")
        chmod_err = None
        try:
            os.chmod(str(path), stat.S_IRWXU)
        except Exception as e:
            chmod_err = e
        # another reason for returning exception:
        # We don't want to interrupt the current line being printed like in slow_delete
        if path.is_dir():
            shutil.rmtree(str(path), ignore_errors=True)  # ignore_errors because of Windows
            try:
                path.unlink(missing_ok=True)  # again, because of Windows
            except IOError:
                pass  # almost definitely because it doesn't exist
        else:
            path.unlink(missing_ok=True)
        logger.debug(f"Permanently deleted {path}")
        return chmod_err

    @classmethod
    def trash(cls, path: PathLike, trash_dir: Optional[PathLike] = None) -> None:
        """
        Trash a file or directory.

        Args:
            path: The path to move to the trash
            trash_dir: If None, uses :meth:`pocketutils.tools.path_tools.PathTools.guess_trash`
        """
        if trash_dir is None:
            trash_dir = PathTools.guess_trash()
        logger.debug(f"Trashing {path} to {trash_dir} ...")
        shutil.move(str(path), str(trash_dir))
        logger.debug(f"Trashed {path} to {trash_dir}")

    @classmethod
    def try_cleanup(cls, path: Path, *, bound: Type[Exception] = PermissionError) -> None:
        """
        Try to delete a file (probably temp file), if it exists, and log any PermissionError.
        """
        path = Path(path)
        # noinspection PyBroadException
        try:
            path.unlink(missing_ok=True)
        except bound:
            logger.error(f"Permission error preventing deleting {path}")

    @classmethod
    def read_lines_file(cls, path: PathLike, ignore_comments: bool = False) -> Sequence[str]:
        """
        Returns a list of lines in the file.
        Optionally skips lines starting with '#' or that only contain whitespace.
        """
        lines = []
        with FilesysTools.open_file(path, "r") as f:
            for line in f.readlines():
                line = line.strip()
                if not ignore_comments or not line.startswith("#") and not len(line.strip()) == 0:
                    lines.append(line)
        return lines

    @classmethod
    def read_properties_file(cls, path: PathLike) -> Mapping[str, str]:
        """
        Reads a .properties file.
        A list of lines with key=value pairs (with an equals sign).
        Lines beginning with # are ignored.
        Each line must contain exactly 1 equals sign.

        Args:
            path: Read the file at this local path

        Returns:
            A dict mapping keys to values, both with surrounding whitespace stripped
        """
        dct = {}
        with FilesysTools.open_file(path, "r") as f:
            for i, line in enumerate(f.readlines()):
                line = line.strip()
                if len(line) == 0 or line.startswith("#"):
                    continue
                if line.count("=") != 1:
                    raise ParsingError(f"Bad line {i} in {path}", resource=path)
                k, v = line.split("=")
                k, v = k.strip(), v.strip()
                if k in dct:
                    raise AlreadyUsedError(f"Duplicate property {k} (line {i})", key=k)
                dct[k] = v
        return dct

    @classmethod
    def write_properties_file(
        cls, properties: Mapping[Any, Any], path: Union[str, PurePath], mode: str = "o"
    ):
        if not OpenMode(mode).write:
            raise ContradictoryRequestError(f"Cannot write text to {path} in mode {mode}")
        with FilesysTools.open_file(path, mode) as f:
            bads = []
            for k, v in properties.items():
                if "=" in k or "=" in v or "\n" in k or "\n" in v:
                    bads.append(k)
                f.write(
                    str(k).replace("=", "--").replace("\n", "\\n")
                    + "="
                    + str(v).replace("=", "--").replace("\n", "\\n")
                    + "\n"
                )
            if 0 < len(bads) <= 10:
                logger.warning(
                    "At least one properties entry contains an equals sign or newline (\\n)."
                    f"These were escaped: {', '.join(bads)}"
                )
            elif len(bads) > 0:
                logger.warning(
                    "At least one properties entry contains an equals sign or newline (\\n),"
                    "which were escaped."
                )

    @classmethod
    def save_json(cls, data: Any, path: PathLike, mode: str = "w") -> None:
        warnings.warn("save_json will be removed; use orjson instead", DeprecationWarning)
        with cls.open_file(path, mode) as f:
            json.dump(data, f, ensure_ascii=False, cls=JsonEncoder)

    @classmethod
    def load_json(cls, path: PathLike):
        warnings.warn("save_json will be removed; use orjson instead", DeprecationWarning)
        return json.loads(Path(path).read_text(encoding="utf8"))

    @classmethod
    def read_any(
        cls, path: PathLike
    ) -> Union[
        str,
        bytes,
        Sequence[str],
        pd.DataFrame,
        Sequence[int],
        Sequence[float],
        Sequence[str],
        Mapping[str, str],
    ]:
        """
        Reads a variety of simple formats based on filename extension.
        Includes '.txt', 'csv', .xml', '.properties', '.json'.
        Also reads '.data' (binary), '.lines' (text lines).
        And formatted lists: '.strings', '.floats', and '.ints' (ex: "[1, 2, 3]").
        """
        path = Path(path)
        ext = path.suffix.lstrip(".")

        def load_list(dtype):
            return [
                dtype(s)
                for s in FilesysTools.read_lines_file(path)[0]
                .replace(" ", "")
                .replace("[", "")
                .replace("]", "")
                .split(",")
            ]

        if ext == "lines":
            return FilesysTools.read_lines_file(path)
        elif ext == "txt":
            return path.read_text("utf-8")
        elif ext == "data":
            return path.read_bytes()
        elif ext == "json":
            return FilesysTools.load_json(path)
        elif ext in ["npy", "npz"]:
            return np.load(str(path), allow_pickle=False)
        elif ext == "properties":
            return FilesysTools.read_properties_file(path)
        elif ext == "csv":
            return pd.read_csv(path)
        elif ext == "ints":
            return load_list(int)
        elif ext == "floats":
            return load_list(float)
        elif ext == "strings":
            return load_list(str)
        elif ext == "xml":
            ElementTree.parse(path).getroot()
        else:
            raise TypeError(f"Did not recognize resource file type for file {path}")

    @classmethod
    @contextmanager
    def open_file(cls, path: PathLike, mode: str):
        """
        Opens a file in a safer way, always using the encoding set in Kale (utf8) by default.
        This avoids the problems of accidentally overwriting, forgetting to set mode, and not setting the encoding.
        Note that the default encoding on open() is not UTF on Windows.
        Raises specific informative errors.
        Cannot set overwrite in append mode.
        """
        path = Path(path)
        mode = OpenMode(mode)
        if mode.write and mode.safe and path.exists():
            raise FileDoesNotExistError(f"Path {path} already exists")
        if not mode.read:
            PathTools.prep_file(path, exist_ok=mode.overwrite or mode.append)
        if mode.gzipped:
            yield gzip.open(path, mode.internal, compresslevel=COMPRESS_LEVEL)
        elif mode.binary:
            yield open(path, mode.internal)
        else:
            yield open(path, mode.internal, encoding=ENCODING)

    @classmethod
    def write_lines(cls, iterable: Iterable[Any], path: PathLike, mode: str = "w") -> int:
        """
        Just writes an iterable line-by-line to a file, using '\n'.
        Makes the parent directory if needed.
        Checks that the iterable is a "true iterable" (not a string or bytes).

        Returns:
            The number of lines written (the same as len(iterable) if iterable has a length)

        Raises:
            FileExistsError: If the path exists and append is False
            PathIsNotFileError: If append is True, and the path exists but is not a file
        """
        path = Path(path)
        mode = OpenMode(mode)
        if not mode.overwrite or mode.binary:
            raise ContradictoryRequestError(f"Wrong mode for writing a text file: {mode}")
        if not cls.is_true_iterable(iterable):
            raise TypeError("Not a true iterable")  # TODO include iterable if small
        PathTools.prep_file(path, exist_ok=mode.overwrite or mode.append)
        n = 0
        with cls.open_file(path, mode) as f:
            for x in iterable:
                f.write(str(x) + "\n")
            n += 1
        return n

    @classmethod
    def hash_hex(cls, x: SupportsBytes, algorithm: str) -> str:
        """
        Returns the hex-encoded hash of the object (converted to bytes).
        """
        m = hashlib.new(algorithm)
        m.update(bytes(x))
        return m.hexdigest()

    @classmethod
    def replace_in_file(cls, path: PathLike, changes: Mapping[str, str]) -> None:
        """
        Uses re.sub repeatedly to modify (AND REPLACE) a file's content.
        """
        path = Path(path)
        data = path.read_text(encoding="utf-8")
        for key, value in changes.items():
            data = regex.sub(key, value, data, flags=regex.V1 | regex.MULTILINE | regex.DOTALL)
        path.write_text(data, encoding="utf-8")

    @classmethod
    def tmppath(cls, path: Optional[PathLike] = None, **kwargs) -> Generator[Path, None, None]:
        """
        Makes a temporary Path. Won't create ``path`` but will delete it at the end.
        If ``path`` is None, will use ``tempfile.mkstemp``.
        """
        if path is None:
            _, path = tempfile.mkstemp()
        try:
            yield Path(path, **kwargs)
        finally:
            Path(path).unlink()

    @classmethod
    def tmpfile(
        cls, path: Optional[PathLike] = None, *, spooled: bool = False, **kwargs
    ) -> Generator[Writeable, None, None]:
        """
        Simple wrapper around tempfile.TemporaryFile, tempfile.NamedTemporaryFile, and tempfile.SpooledTemporaryFile.
        """
        if spooled:
            with tempfile.SpooledTemporaryFile(**kwargs) as x:
                yield x
        elif path is None:
            with tempfile.TemporaryFile(**kwargs) as x:
                yield x
        else:
            with tempfile.NamedTemporaryFile(str(path), **kwargs) as x:
                yield x

    @classmethod
    def tmpdir(cls, **kwargs) -> Generator[Path, None, None]:
        with tempfile.TemporaryDirectory(**kwargs) as x:
            yield Path(x)


__all__ = ["FilesysTools"]



from typing import Optional, SupportsFloat, SupportsInt, TypeVar

import numpy as np

from pocketutils.tools.base_tools import BaseTools

V = TypeVar("V")


class NumericTools(BaseTools):
    @classmethod
    def fnone(cls, f: Optional[SupportsFloat]) -> float:
        return None if f is None else float(f)

    @classmethod
    def iroundopt(cls, f: Optional[SupportsInt]) -> int:
        # noinspection PyTypeChecker
        return None if f is None else int(np.round(f))

    @classmethod
    def iceilopt(cls, f: Optional[SupportsInt]) -> int:
        # noinspection PyTypeChecker
        return None if f is None else int(np.ceil(f))

    @classmethod
    def iflooropt(cls, f: Optional[SupportsInt]) -> int:
        # noinspection PyTypeChecker
        return None if f is None else int(np.floor(f))

    @classmethod
    def iround(cls, f: Optional[SupportsInt]) -> int:
        # noinspection PyTypeChecker
        return None if f is None else int(np.round(f))

    @classmethod
    def iceil(cls, f: SupportsFloat) -> int:
        """
        Fixes np.ceil to return an Python integer rather than a Numpy float.
            f: A Python or Numpy float, or something else that defines __float__

        Returns:
            An integer of the ceiling
        """
        # noinspection PyTypeChecker
        return int(np.ceil(f))

    @classmethod
    def ifloor(cls, f: SupportsFloat) -> int:
        """
        Fixes np.floor to return an Python integer rather than a Numpy float.
            f: A Python or Numpy float, or something else that defines __float__

        Returns:
            An integer of the ceiling
        """
        # noinspection PyTypeChecker
        return int(np.floor(f))

    @classmethod
    def imin(cls, *f) -> int:
        return int(np.min(f))

    @classmethod
    def imax(cls, *f) -> int:
        return int(np.max(f))

    @classmethod
    def slice_bounded(cls, arr: np.array, i: Optional[int], j: Optional[int]) -> np.array:
        """
        Slices ``arr[max(i,0), min(j, len(arr))``.
        Converts ``i`` and ``j`` to int.
        """
        if i is None:
            i = 0
        if j is None:
            j = len(arr)
        if i < 0:
            i = len(arr) - i
        if j < 0:
            j = len(arr) - j
        start, stop = NumericTools.imax(0, i), NumericTools.imin(len(arr), j)
        return arr[start:stop]


__all__ = ["NumericTools"]



from pocketutils.tools.call_tools import CallTools
from pocketutils.tools.common_tools import CommonTools
from pocketutils.tools.console_tools import ConsoleTools
from pocketutils.tools.filesys_tools import FilesysTools
from pocketutils.tools.numeric_tools import NumericTools
from pocketutils.tools.path_tools import PathTools
from pocketutils.tools.program_tools import ProgramTools
from pocketutils.tools.string_tools import StringTools
from pocketutils.tools.unit_tools import UnitTools


class Tools(
    CallTools,
    CommonTools,
    ConsoleTools,
    NumericTools,
    PathTools,
    ProgramTools,
    StringTools,
    FilesysTools,
    UnitTools,
):
    """
    A collection of utility static functions.
    """


__all__ = ["Tools"]



"""
Loguru logging extension that, configurably:
- redirects built-in logging to your loguru logger
- remembers the handlers added
- auto-detects compression and serialization from filenames
- includes extras in non-serializing handlers
- has a few alternative (possibly better) choices for colors, icons, and levels
- will complain when you do really stupid things
- has a convenient notation for configuring from a CLI
  (e.g. ``--stderr debug --log :INFO:run.log.gz``)
- mandes utf-8
"""
from __future__ import annotations

import logging
import os
import sys
from dataclasses import dataclass
from inspect import cleandoc
from pathlib import Path
from typing import AbstractSet, Any, Callable, Generic, Mapping, Optional, TextIO, TypeVar, Union

# noinspection PyProtectedMember
import loguru._defaults as _defaults
import regex
from loguru import logger

# noinspection PyProtectedMember
from loguru._logger import Logger

from pocketutils.core.exceptions import IllegalStateError, XValueError

Formatter = Union[str, Callable[[Mapping[str, Any]], str]]
_FMT = cleandoc(
    r"""
    <bold>{time:YYYY-MM-DD HH:mm:ss.SSS}</bold> |
    <level>{level: <8}</level> |
    <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>
    — <level>{message}[EXTRA]</level>
    {exception}
    """
).replace("\n", " ")


class FormatFactory:
    @classmethod
    def with_extras(
        cls,
        *,
        fmt: str = _FMT,
        sep: str = "; ",
        eq_sign: str = " ",
    ) -> Callable[[Mapping[str, Any]], str]:
        def FMT(record: Mapping[str, Any]) -> str:
            extra = sep.join([e + eq_sign + "{extra[" + e + "]}" for e in record["extra"].keys()])
            if len(extra) > 0:
                extra = f" [ {extra} ]"
            return fmt.replace("[EXTRA]", extra) + os.linesep

        return FMT

    @classmethod
    def plain(cls, *, fmt: str = _FMT) -> Callable[[Mapping[str, Any]], str]:
        def FMT(record: Mapping[str, Any]) -> str:
            return fmt.replace("[EXTRA]", "")

        return FMT


class _SENTINEL:
    pass


T = TypeVar("T", covariant=True, bound=Logger)

_LOGGER_ARG_PATTERN = regex.compile(r"(?:([a-zA-Z]+):)?(.*)", flags=regex.V1)
log_compressions = {
    ".xz",
    ".lzma",
    ".gz",
    ".zip",
    ".bz2",
    ".tar",
    ".tar.gz",
    ".tar.bz2",
    ".tar.xz",
}
valid_log_suffixes = {
    *{f".log{c}" for c in log_compressions},
    *{f".txt{c}" for c in log_compressions},
    *{f".json{c}" for c in log_compressions},
}


class FancyLoguruDefaults:
    levels_built_in = dict(
        TRACE=_defaults.LOGURU_TRACE_NO,
        DEBUG=_defaults.LOGURU_DEBUG_NO,
        INFO=_defaults.LOGURU_INFO_NO,
        WARNING=_defaults.LOGURU_WARNING_NO,
        ERROR=_defaults.LOGURU_ERROR_NO,
        CRITICAL=_defaults.LOGURU_CRITICAL_NO,
    )

    # the levels for caution and notice are DEFINED here
    # trace and success must match loguru's
    # and the rest must match logging's
    # note that most of these alternate between informative and problematic
    # i.e. info (ok), caution (bad), success (ok), warning (bad), notice (ok), error (bad)
    levels_extended = {
        **levels_built_in,
        **dict(
            CAUTION=23,
            SUCCESS=25,
            NOTICE=35,
        ),
    }

    colors_built_in = dict(
        TRACE=_defaults.LOGURU_TRACE_COLOR,
        DEBUG=_defaults.LOGURU_DEBUG_COLOR,
        INFO=_defaults.LOGURU_INFO_COLOR,
        SUCCESS=_defaults.LOGURU_SUCCESS_COLOR,
        WARNING=_defaults.LOGURU_WARNING_COLOR,
        ERROR=_defaults.LOGURU_ERROR_COLOR,
        CRITICAL=_defaults.LOGURU_CRITICAL_COLOR,
    )

    colors_extended = dict(
        **colors_built_in,
        CAUTION=_defaults.LOGURU_WARNING_COLOR,
        NOTICE=_defaults.LOGURU_INFO_COLOR,
    )

    colors_red_green_safe = dict(
        TRACE="<dim>",
        DEBUG="<dim>",
        INFO="<bold>",
        CAUTION="<yellow>",
        SUCCESS="<blue>",
        WARNING="<yellow>",
        NOTICE="<blue>",
        ERROR="<red>",
        CRITICAL="<red>",
    )

    icons_built_in = dict(
        TRACE=_defaults.LOGURU_TRACE_ICON,
        DEBUG=_defaults.LOGURU_DEBUG_ICON,
        INFO=_defaults.LOGURU_INFO_ICON,
        SUCCESS=_defaults.LOGURU_SUCCESS_ICON,
        WARNING=_defaults.LOGURU_WARNING_ICON,
        ERROR=_defaults.LOGURU_ERROR_ICON,
        CRITICAL=_defaults.LOGURU_CRITICAL_ICON,
    )

    icons_extended = dict(
        **icons_built_in,
        CAUTION="⚐",
        NOTICE="★",
    )

    level: str = "INFO"

    fmt = FormatFactory.with_extras()

    aliases = dict(NONE=None, NO=None, OFF=None, VERBOSE="INFO", QUIET="ERROR")


@dataclass(frozen=True, repr=True, order=True)
class HandlerInfo:
    hid: int
    path: Optional[Path]
    level: Optional[str]
    fmt: Formatter


@dataclass(frozen=False, repr=True)
class _HandlerInfo:
    hid: int
    sink: Any
    level: Optional[int]
    fmt: Formatter


@dataclass(frozen=True, repr=True, order=True)
class LogSinkInfo:
    path: Path
    base: Path
    suffix: str
    serialize: bool
    compression: Optional[str]

    @classmethod
    def guess(cls, path: Union[str, Path]) -> LogSinkInfo:
        path = Path(path)
        base, compression = path.name, None
        for c in log_compressions:
            if path.name.endswith(c):
                base, compression = path.name[: -len(c)], c
        if not [base.endswith(s) for s in [".json", ".log", ".txt"]]:
            raise XValueError(
                f"Log filename {path.name} is not .json, .log, .txt, or a compressed variant"
            )
        return LogSinkInfo(
            path=path,
            base=path.parent / base,
            suffix=compression,
            serialize=base.endswith(".json"),
            compression=compression,
        )


class InterceptHandler(logging.Handler):
    """
    Redirects standard logging to loguru.
    """

    def emit(self, record):
        # Get corresponding Loguru level if it exists
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        # Find caller from where originated the logged message
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1
        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())


class FancyLoguru(Generic[T]):
    def __init__(self, log: T):
        self._levels = dict(FancyLoguruDefaults.levels_built_in)
        self._logger = log
        self._main = None
        self._paths = {}
        self._aliases = dict(FancyLoguruDefaults.aliases)

    @property
    def logger(self) -> T:
        return self._logger

    @property
    def levels(self) -> Mapping[str, int]:
        return dict(self._levels)

    @property
    def aliases(self) -> Mapping[str, str]:
        return self._aliases

    @property
    def main(self) -> Optional[HandlerInfo]:
        if self._main is None:
            return None
        return HandlerInfo(
            hid=self._main.hid, level=self._main.level, fmt=self._main.fmt, path=None
        )

    @property
    def paths(self) -> AbstractSet[HandlerInfo]:
        if self._main is None:
            return set()
        return {HandlerInfo(hid=h.hid, level=h.level, fmt=h.fmt, path=h.sink) for h in self._paths}

    def config_levels(
        self,
        *,
        levels: Mapping[str, int] = _SENTINEL,
        colors: Mapping[str, str] = _SENTINEL,
        icons: Mapping[str, str] = _SENTINEL,
        aliases: Mapping[str, str] = _SENTINEL,
    ) -> __qualname__:
        levels = FancyLoguruDefaults.levels_extended if levels is _SENTINEL else levels
        colors = FancyLoguruDefaults.colors_extended if colors is _SENTINEL else colors
        icons = FancyLoguruDefaults.icons_extended if icons is _SENTINEL else icons
        aliases = FancyLoguruDefaults.aliases if aliases is _SENTINEL else aliases
        for k, v in levels.items():
            self.config_level(k, v, color=colors.get(k, _SENTINEL), icon=icons.get(k, _SENTINEL))
        self._aliases = dict(aliases)
        return self

    def init(
        self,
        *,
        level: str = FancyLoguruDefaults.level,
        sink=sys.stderr,
        fmt: Formatter = FancyLoguruDefaults.fmt,
        intercept: bool = True,
    ) -> __qualname__:
        """
        Sets an initial configuration.
        """
        if intercept:
            # noinspection PyArgumentList
            logging.basicConfig(handlers=[InterceptHandler()], level=0, encoding="utf-8")
        self.logger.remove(None)  # get rid of the built-in handler
        self.config_main(level=level, sink=sink, fmt=fmt)
        return self

    def config_level(
        self,
        name: str,
        level: int,
        *,
        color: Union[None, str, _SENTINEL] = _SENTINEL,
        icon: Union[None, str, _SENTINEL] = _SENTINEL,
        replace: bool = True,
    ) -> __qualname__:
        try:
            data = logger.level(name)
        except ValueError:
            data = None
        if data is None:
            logger.level(
                name,
                no=level,
                color=None if color is _SENTINEL else color,
                icon=None if icon is _SENTINEL else icon,
            )
        elif replace:
            if level != data.no:  # loguru doesn't check whether they're eq; it just errors
                raise IllegalStateError(f"Cannot set level={level}!={data.no} for {name}")
            logger.level(
                name,
                color=data.color if color is _SENTINEL else color,
                icon=data.icon if icon is _SENTINEL else icon,
            )
        return self

    def config_main(
        self,
        *,
        sink: TextIO = _SENTINEL,
        level: Optional[str] = _SENTINEL,
        fmt: Formatter = _SENTINEL,
    ) -> __qualname__:
        """
        Sets the logging level for the main handler (normally stderr).
        """
        if level is not None and level is not _SENTINEL:
            level = level.upper()
        if self._main is None:
            self._main = _HandlerInfo(
                hid=-1,
                sink=sys.stderr,
                level=self._levels[FancyLoguruDefaults.level],
                fmt=FancyLoguruDefaults.fmt,
            )
        else:
            try:
                logger.remove(self._main.hid)
            except ValueError:
                logger.error(f"Cannot remove handler {self._main.hid}")
        self._main.level = self._main.level if level is _SENTINEL else level
        self._main.sink = self._main.sink if sink is _SENTINEL else sink
        self._main.fmt = self._main.fmt if fmt is _SENTINEL else fmt
        self._main.hid = logger.add(self._main.sink, level=self._main.level, format=self._main.fmt)
        return self

    def remove_path(self, path: Path) -> __qualname__:
        for k, h in self._paths.items():
            if h.path.resolve() == path.resolve():
                h.level = None
                try:
                    logger.remove(k)
                except ValueError:
                    logger.error(f"Cannot remove handler {k} to {path}")

    def add_path(
        self,
        path: Path,
        level: str = FancyLoguruDefaults.level,
        *,
        fmt: str = FancyLoguruDefaults.fmt,
    ) -> __qualname__:
        level = level.upper()
        ell = self._levels[level]
        info = LogSinkInfo.guess(path)
        x = logger.add(
            str(info.base),
            format=fmt,
            level=level,
            compression=info.compression,
            serialize=info.serialize,
            backtrace=True,
            diagnose=True,
            enqueue=True,
            encoding="utf-8",
        )
        self._paths[x] = _HandlerInfo(hid=x, sink=info.base, level=ell, fmt=fmt)
        return self

    def from_cli(
        self,
        path: Union[None, str, Path] = None,
        main: Optional[str] = FancyLoguruDefaults.level,
    ) -> __qualname__:
        """
        This function controls logging set via command-line.

        Args:
            main: The level for stderr
            path: If set, the path to a file. Can be prefixed with ``:level:`` to set the level
                  (e.g. ``:INFO:mandos-run.log.gz``). Can serialize to JSON if .json is used
                  instead of .log or .txt.
        """
        if main is None:
            main = FancyLoguruDefaults.level
        main = self._aliases.get(main.upper(), main.upper())
        if main not in FancyLoguruDefaults.levels_extended:
            _permitted = ", ".join(
                [*FancyLoguruDefaults.levels_extended, *FancyLoguruDefaults.aliases.keys()]
            )
            raise XValueError(f"{main.lower()} not a permitted log level (allowed: {_permitted}")
        self.config_main(level=main)
        if path is not None or len(str(path)) == 0:
            match = _LOGGER_ARG_PATTERN.match(str(path))
            path_level = "DEBUG" if match.group(1) is None else match.group(1)
            path = Path(match.group(2))
            self.add_path(path, path_level)
            self.logger.info(f"Added logger to {path} at level {path_level}")
        self.logger.info(f"Set main log level to {main}")
        return self

    __call__ = from_cli


class FancyLoguruExtras:
    @classmethod
    def force_streams_to_utf8(cls) -> None:
        # we warn the user about this in the docs!
        sys.stderr.reconfigure(encoding="utf-8")
        sys.stdout.reconfigure(encoding="utf-8")
        sys.stdin.reconfigure(encoding="utf-8")


__all__ = ["FancyLoguruDefaults", "FancyLoguru", "FancyLoguruExtras", "HandlerInfo"]



from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional, Union

from pocketutils.core import LazyWrap, PathLike

logger = logging.getLogger("pocketutils")


class MagicTemplate:
    @classmethod
    def from_path(cls, path: PathLike, *, prefix: str = "${{", suffix: str = "}}") -> __qualname__:
        return MagicTemplate(
            reader=lambda: Path(path).read_text(encoding="utf8"), prefix=prefix, suffix=suffix
        )

    @classmethod
    def from_text(cls, text: str, *, prefix: str = "${{", suffix: str = "}}") -> __qualname__:
        return MagicTemplate(reader=lambda: text, prefix=prefix, suffix=suffix)

    def __init__(self, *, reader: Callable[[], str], prefix: str = "${{", suffix: str = "}}"):
        self._reader = reader
        self._entries = {}
        self._prefix, self._suffix = prefix, suffix

    def add(self, key: str, value: Union[Any, Callable[[str], str]]) -> __qualname__:
        self._entries[key] = value
        return self

    def add_version(self, semantic_version: str) -> __qualname__:
        self._entries.update(
            {
                "version": semantic_version,
                "major": semantic_version.split(".")[0],
                "minor": semantic_version.split(".")[1] if semantic_version.count(".") > 0 else "-",
                "patch": semantic_version.split(".")[2] if semantic_version.count(".") > 1 else "-",
            }
        )
        return self

    def add_datetime(self, at: Optional[datetime] = None) -> __qualname__:
        if at is None:
            now = LazyWrap.new_type("datetime", datetime.now)()
        else:
            now = LazyWrap.new_type("datetime", lambda: at)()
        self._entries.update(
            {
                "year": lambda _: str(now.get().year),
                "month": lambda _: str(now.get().month),
                "day": lambda _: str(now.get().day),
                "hour": lambda _: str(now.get().hour),
                "minute": lambda _: str(now.get().minute),
                "second": lambda _: str(now.get().second),
                "datestr": lambda _: str(now.get().date()),
                "timestr": lambda _: str(now.get().time()),
                "datetuple": lambda _: str((now.get().year, now.get().month, now.get().day)),
                "datetime": lambda _: str(
                    (
                        now.get().year,
                        now.get().month,
                        now.get().day,
                        now.get().hour,
                        now.get().minute,
                        now.get().second,
                    )
                ),
            }
        )
        return self

    def register_magic(self, name: str, shell=None) -> None:
        if shell is None:
            from IPython import get_ipython

            shell = get_ipython()
        shell.register_magic_function(self._fill, magic_kind="line_cell", magic_name=name)

    def parse(self, line: str = "") -> str:
        return self.__replace(self._reader(), line)

    def _fill(self, line, shell=None):
        if shell is None:
            from IPython import get_ipython

            shell = get_ipython()
        text = self.__replace(self._reader(), line)
        shell.set_next_input(text, replace=True)

    def __replace(self, r: str, line: str):
        for k, v in self._entries.items():
            k = self._prefix + k + self._suffix
            if k in r:
                try:
                    if callable(v):
                        r = r.replace(k, v(line))
                    else:
                        r = r.replace(k, str(v))
                except Exception:
                    logger.error(f"Failed replacing {k}")
                    raise
        return r


__all__ = ["MagicTemplate"]
